import os
from typing import Optional, Union, List, Dict
import logging
from datetime import datetime
import xml.etree.ElementTree as ET

from ..xml_parser import XMLParser
from ..gpx_elements import Bounds, Copyright, Email, Extensions, Gpx, Link, Metadata, Person, Point, PointSegment, Route, TrackSegment, Track, WayPoint

class KMLParser(XMLParser):
    """
    KML file parser.
    """

    def __init__(
            self,
            file_path: Optional[str] = None,
            check_xml_schemas: bool = True,
            xml_extensions_schemas: bool = False) -> None:
        """
        Initialize KMLParser instance.

        Args:
            file_path (str, optional): Path to the file to parse. Defaults to None.
            check_xml_schemas (bool, optional): Toggle schema verification during parsing. Defaults to True.
            xml_extensions_schemas (bool, optional): Toggle extensions schema verificaton durign parsing. Requires internet connection and is not guaranted to work. Defaults to False.
        """
        if not file_path.endswith(".kml"):
            return
        super().__init__(file_path,
                         {"opengis": "http://www.opengis.net/kml/2.2"},
                         check_xml_schemas,
                         xml_extensions_schemas)
        
        if self.file_path is not None and os.path.exists(self.file_path):
            self.parse()
        else:
            logging.warning("File path does not exist")

    def find_precisions(self):
        """
        Find decimal precision of any type of value in a KML file (latitude, elevation...).
        """
        # Point
        documents = self.xml_root.findall("opengis:Document", self.name_space)
        placemarks = documents[0].findall("opengis:Placemark", self.name_space)
        linestrings = placemarks[0].findall("opengis:LineString", self.name_space)
        coordinates = self.find_text(linestrings[0], "opengis:coordinates")

        coordinates = coordinates.replace("\n", "").replace("\t", "")
        if coordinates[-1] == " ":
            coordinates = coordinates[:-1]
        coordinates = coordinates.split(" ")
        coordinates = coordinates[0].split(",")

        self.precisions["lat_lon"] = self.find_precision(coordinates[0])
        self.precisions["elevation"] = self.find_precision(coordinates[2])

    # def parse_linestring(self, linestring) -> List[str]:
    #     """
    #     Parse LineString element from KML file.

    #     Args:
    #         placemark (xml.etree.ElementTree.Element): Parsed LineString element.

    #     Returns:
    #         List[str]: Informations contained in the LineString element (strings of coordinates).
    #     """
    #     if linestring is None:
    #         return None

    #     linestrings_data = []
    #     coordinatess = linestring.findall("opengis:coordinates", self.name_space)
    #     for coordinates in coordinatess:
    #         linestrings_data.append(coordinates.text)

    #     return linestrings_data

    def parse_placemark(self, placemark) -> Union[Dict, None]:
        """
        Parse Placemark element from KML file.

        Args:
            placemark (xml.etree.ElementTree.Element): Parsed Placemark element.

        Returns:
            Union[List[Dict], None]: Informations contained in the Placemark element (the name of the Placemark (str) and the contents of the LineString (List[str])).
        """
        if placemark is None:
            return None
        
        placemark_data = {}
        
        placemark_data["name"] = self.find_text(placemark, "opengis:name")

        placemark_data["linestrings_data"] = []
        linestrings = placemark.findall("opengis:LineString", self.name_space)
        for linestring in linestrings:
            placemark_data["linestrings_data"].append(self.find_text(linestring, "opengis:coordinates"))

        return placemark_data

    def parse_document(self, document) -> Union[List[Dict], None]:
        """
        Parse Document element from KML file.

        Args:
            document (xml.etree.ElementTree.Element): Parsed Document element.

        Returns:
            Union[List[Dict], None]: Informations related to Placemark elements contained in
            the Document element.
        """
        if document is None:
            return None
        
        # name = self.find_text(document, "opengis:name")

        placemmarks_data = []
        placemarks = document.findall("opengis:Placemark", self.name_space)
        for placemark in placemarks:
            placemmarks_data.append(self.parse_placemark(placemark))

        return placemmarks_data

    def parse_root_document(self):
        """
        Parse Document elements from KML file.
        """
        documents = self.xml_root.findall("opengis:Document", self.name_space)
        for document in documents:
            placemarks_data = self.parse_document(document)

            if len(placemarks_data) == 1:
                placemark_data = placemarks_data[0]
                linestrings_data = placemark_data["linestrings_data"]
                trkseg = []
                for coordinates in linestrings_data:
                    coordinates = coordinates.replace("\n", "").replace("\t", "")
                    if coordinates[-1] == " ":
                        coordinates = coordinates[:-1]
                    coordinates = coordinates.split(" ")
                    trkpt = []
                    for point_coord in coordinates:
                        point_coord = point_coord.split(",")
                        trkpt.append(WayPoint(tag="trkpt",
                                              lat=float(point_coord[1]),
                                              lon=float(point_coord[0]),
                                              ele=float(point_coord[2])))
                    trkseg.append(TrackSegment(trkpt=trkpt))

                tracks = [Track(name=placemark_data["name"], trkseg=trkseg)]
                self.gpx.tracks = tracks
            else:
                logging.error("Oops, not yet implemented...")

    def add_properties(self):
        self.gpx.creator = "ezGPX"
        self.gpx.xmlns = "http://www.topografix.com/GPX/1/1"
        self.gpx.version = "1.1"
        self.gpx.xmlns_xsi = "http://www.w3.org/2001/XMLSchema-instance"
        self.gpx.xsi_schema_location = ["http://www.topografix.com/GPX/1/1", "http://www.topografix.com/GPX/1/1/gpx.xsd"]

    def parse(self) -> Gpx:
        """
        Parse KML file.

        Returns:
            Gpx: Gpx instance.
        """
        # Parse KML file
        try:
            self.xml_tree = ET.parse(self.file_path)
            self.xml_root = self.xml_tree.getroot()
        except Exception as err:
            logging.exception(f"Unexpected {err}, {type(err)}.\nUnable to parse KML file.")
            raise

        # Add properties
        self.add_properties()

        # Find precisions
        self.find_precisions()

        # Check XML schemas
        self.check_xml_schemas()

        # Parse Document
        try:
            self.parse_root_document()
        except:
            logging.error("Unable to parse tracks in GPX file.")
            raise

        logging.debug("Parsing complete!!")
        return self.gpx


from typing import Optional, Union
import logging
from datetime import datetime
import xml.etree.ElementTree as ET

from ..parser import Parser

class XMLParser(Parser):
    """
    XML File parser.
    """

    def __init__(
            self,
            file_path: Optional[str] = None,
            name_space: Optional[dict] = None,
            xml_schema: bool = True,
            xml_extensions_schemas: bool = False) -> None:
        """
        Initialize XML Parser instance.

        Args:
            file_path (str, optional): Path to the file to parse. Defaults to None.
            name_space (dict, optional): File XML name spaces.
            check_xml_schemas (bool, optional): Toggle schema verification during parsing. Defaults to True.
            xml_extensions_schemas (bool, optional): Toggle extensions schema verificaton durign parsing. Requires internet connection and is not guaranted to work. Defaults to False.
        """
        super().__init__(file_path)
        
        self.name_space: dict = None
        if name_space is None:
            self.name_space = {}
        else:
            self.name_space = name_space

        self.xml_schema: bool = xml_schema
        self.xml_extensions_schemas: bool = xml_extensions_schemas

        self.xml_tree: ET.ElementTree = None
        self.xml_root: ET.Element = None

    def get_text(self, element, sub_element: str) -> Union[str, None]:
        """
        Get text from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[str, None]: Text from sub-element.
        """
        try:
            text_ = element.get(sub_element)
        except:
            logging.debug(f"{element} has no attribute {sub_element}.")
            text_ = None
        return text_
    
    def get_int(self, element, sub_element: str) -> Union[int, None]:
        """
        Get integer value from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[int, None]: Integer value from sub-element.
        """
        try:
            int_ = int(element.get(sub_element))
        except:
            logging.debug(f"{element} has no attribute {sub_element}.")
            int_ = None
        return int_
    
    def get_float(self, element, sub_element: str) -> Union[float, None]:
        """
        Get floating point value from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[float, None]: Floating point value from sub-element.
        """
        try:
            float_ = float(element.get(sub_element))
        except:
            logging.debug(f"{element} has no attribute {sub_element}.")
            float_ = None
        return float_
    
    def find_text(self, element, sub_element: str) -> Union[str, None]:
        """
        Find text from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[str, None]: Text from sub-element.
        """
        try:
            text_ = element.find(sub_element, self.name_space).text
        except:
            text_ = None
            logging.debug(f"{element} has no attribute {sub_element}.")
        return text_
    
    def find_int(self, element, sub_element: str) -> Union[int, None]:
        """
        Find integer value from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[int, None]: Integer value from sub-element.
        """
        try:
            int_ = int(element.find(sub_element, self.name_space).text)
        except:
            int_ = None
            logging.debug(f"{element} has no attribute {sub_element}.")
        return int_
    
    def find_float(self, element, sub_element: str) -> Union[float, None]:
        """
        Find float point value from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[float, None]: Floating point value from sub-element.
        """
        try:
            float_ = float(element.find(sub_element, self.name_space).text)
        except:
            float_ = None
            logging.debug(f"{element} has no attribute {sub_element}.")
        return float_
    
    def find_time(self, element, sub_element: str) -> Union[datetime, None]:
        """
        Find time value from sub-element.

        Args:
            element (xml.etree.ElementTree.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            Union[datetime, None]: Floating point value from sub-element.
        """
        try:
            time_ = datetime.strptime(element.find(sub_element, self.name_space).text, self.time_format)
        except:
            time_ = None
            logging.debug(f"{element} has no attribute {sub_element}.")
        return time_
    
    def check_xml_schemas(self):
        """
        Check XML schemas during parsing.
        """
        # Check XML schema
        if self.xml_schema:
            if not self.gpx.check_xml_schema(self.file_path):
                logging.error("Invalid GPX file (does not follow XML schema).")
                raise

        # Check XML extension schemas
        if self.xml_extensions_schemas:
            if not self.gpx.check_xml_extensions_schemas(self.file_path):
                logging.error("Invalid GPX file (does not follow XML extensions schemas).")
                raise



class Email():
    """
    emailType element in GPX file.
    """

    def __init__(
            self,
            tag: str = "email",
            id: str = None,
            domain: str = None) -> None:
        """
        Initialize Email instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "email"
        id : str, optional
            Identifier, by default None
        domain : str, optional
            Domain, by default None
        """
        self.tag: str = tag
        self.id: str = id
        self.domain: str = domain



class Link():
    """
    linkType element in GPX file.
    """

    def __init__(
            self,
            tag: str = "link",
            href: str = None,
            text: str = None,
            type: str = None) -> None:
        """
        Initialize Link instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "link"
        href : str, optional
            Hyper reference, by default None
        text : str, optional
            Text, by default None
        type : str, optional
            Type, by default None
        """
        self.tag: str = tag
        self.href: str = href
        self.text: str = text
        self.type: str = type
        



from typing import List
import logging
from math import degrees

from .distance import EARTH_RADIUS, perpendicular_distance

def ramer_douglas_peucker(points: List, epsilon: float = degrees(2/EARTH_RADIUS)):
    """
    Simplify a curve using the Ramer-Douglas-Peucker algorithm.
    Source: https://en.wikipedia.org/wiki/Ramer%E2%80%93Douglas%E2%80%93Peucker_algorithm

    Parameters
    ----------
    points : List
        List of points defining the track to simplify.
    epsilon : float, optional
        Ramer-Douglas-Peucker threshold distance (higher value means more simplifications), by default degrees(2/EARTH_RADIUS) (ie: the angle corresponding to a distance of 2 meters at the surface of the earth).

    Returns
    -------
    List
        List of points defining the simplified track.
    """
    # Find the point with the maximum distance
    d_max = 0
    i_max = 0

    start_point = points[0]
    end_point = points[len(points)-1]

    for i in range(1, len(points)-1):
        d = perpendicular_distance(start_point, end_point, points[i])
        if d > d_max:
            d_max = d
            i_max = i

    # logging.info(f"i_max = {i_max} | d_max = {d_max}")

    result = []

    # If max distance is greater than epsilon, recursively simplify
    if d_max > epsilon:
        # Recursive call
        result_1 = ramer_douglas_peucker(points[0:i_max+1], epsilon)
        result_2 = ramer_douglas_peucker(points[i_max: len(points)], epsilon)

        # Build result list
        result = result_1 + result_2[1:]
    else:
        result = [start_point, end_point]

    return result



class Copyright():
    """
    copyrightType element in GPX file.
    """

    def __init__(
            self,
            tag: str = "copyright",
            author: str = None,
            year: int = None,
            licence: str = None) -> None:
        """
        Initialize Copyright instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "copyright"
        author : str, optional
            Author, by default None
        year : int, optional
            Year, by default None
        licence : str, optional
            Licence, by default None
        """
        self.tag: str = tag
        self.author: str = author
        self.year: int = year
        self.licence: str = licence
        


import os
from typing import Optional, Union, Tuple
import logging
import xml.etree.ElementTree as ET
from datetime import datetime

from ..gpx_elements import Gpx
from ..parser import DEFAULT_PRECISION, DEFAULT_TIME_FORMAT

class Writer():
    """
    File writer.
    """

    def __init__(
            self,
            gpx: Gpx = None,
            file_path: str = None) -> None:
        """
        Initialize Writer instance.

        Parameters
        ----------
        gpx : Gpx, optional
            Gpx instance to write, by default None
        file_path : str, optional
            Path to the file to write, by default None
        """
        self.gpx: Gpx = gpx
        self.file_path: str = file_path

    def setIfNotNone(self, element: ET.Element, field: str, value):
        """
        _summary_

        Parameters
        ----------
        element : ET.Element
            _description_
        field : str
            _description_
        value : _type_
            _description_
        """
        if value is not None:
            element.set(field, value)

    def add_subelement(self, element: ET.Element, sub_element: str, text: str) -> Tuple[ET.Element, Union[ET.Element, None]]:
        """
        Add sub-element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            sub_element (str): GPX sub-element.
            text (str): GPX sub-element text.

        Returns:
            Tuple[xml.etree.ElementTree.Element, Union[xml.etree.ElementTree.Element, None]]: GPX element and GPX sub-element (if not None).
        """
        sub_element_ = None
        if text is not None:
            sub_element_ = ET.SubElement(element, sub_element)
            sub_element_.text = text
        return element, sub_element_
    
    def add_subelement_number(self, element: ET.Element, sub_element: str, number: Union[int, float], precision: int = DEFAULT_PRECISION) -> Tuple[ET.Element, Union[ET.Element, None]]:
        """
        Add sub-element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            sub_element (str): GPX sub-element.
            number (Union[int, float], optional): GPX sub-element value.
            precision (int, optional): Precision. Defaults to DEFAULT_PRECISION.

        Returns:
            Tuple[xml.etree.ElementTree.Element, Union[xml.etree.ElementTree.Element, None]]: GPX element and GPX sub-element (if not None).
        """
        sub_element_ = None
        if number is not None:
            sub_element_ = ET.SubElement(element, sub_element)
            if type(number) is int:
                sub_element_.text = str(number)
            elif type(number) is float:
                sub_element_.text = "{:.{}f}".format(number, precision)
            else:
                logging.error("Invalid number type.")
        return element, sub_element_
    
    def add_subelement_time(self, element: ET.Element, sub_element: str, time: datetime, format: str = DEFAULT_TIME_FORMAT) -> ET.Element:
        """
        Add sub-element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            sub_element (str): GPX sub-element.
            time (datetime, optional): GPX sub-element value.
            format (str, optional): Format. Defaults to DEFAULT_TIME_FORMAT.

        Returns:
            Tuple[xml.etree.ElementTree.Element, Union[xml.etree.ElementTree.Element, None]]: GPX element and GPX sub-element (if not None).
        """
        sub_element_ = None
        if time is not None:
            sub_element_ = ET.SubElement(element, sub_element)
            try:
                sub_element_.text = time.strftime(format)
            except:
                logging.error("Invalid time format.")
        return element, sub_element_
    
    def check_xml_schemas(
            self,
            xml_schema: bool = False,
            xml_extensions_schemas: bool = False) -> bool:
        """
        Check XML schemas after writting.

        Parameters
        ----------
        xml_schema : bool, optional
            Toggle XML schema verification, by default False.
        xml_extensions_schemas : bool, optional
            Toggle XML extensions schemas verification, by default False.

        Returns
        -------
        bool
            True if the written file follows all verified schemas.
        """
        # Check XML schema
        if xml_schema:
            if not self.gpx.check_xml_schema(self.file_path):
                logging.error("Invalid GPX file (does not follow XML schema).")
                return False

        # Check XML extension schemas
        if xml_extensions_schemas:
            if not self.gpx.check_xml_extensions_schemas(self.file_path):
                logging.error("Invalid GPX file (does not follow XML extensions schemas).")
                return False
            
        return True


import os
from importlib_resources import files
from typing import Union, List, Tuple
import logging
import xmlschema
import pandas as pd
from datetime import datetime, timezone

from .metadata import Metadata
from .way_point import WayPoint
from .route import Route
from .track import Track
from .extensions import Extensions
from ..utils import haversine_distance, ramer_douglas_peucker

class Gpx():
    """
    gpxType element in GPX file.
    """

    def __init__(
            self,
            tag: str = "gpx",
            creator: str = None,
            xmlns: str = None,
            version: str = None,
            xmlns_xsi: str = None,
            xsi_schema_location: List[str] = None,
            xmlns_gpxtpx: str = None,
            xmlns_gpxx: str = None,
            xmlns_gpxtrk: str = None,
            xmlns_wptx1: str = None,
            metadata: Metadata = None,
            wpt: List[WayPoint] = None,
            rte: List[Route] = None,
            tracks: List[Track] = None,
            extensions: Extensions = None) -> None:
        """
        Initialize Gpx instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "gpx"
        creator : str, optional
            Creator, by default None
        xmlns : str, optional
            XML xmlns, by default None
        version : str, optional
            Version, by default None
        xmlns_xsi : str, optional
            XML xmlns_xsi, by default None
        xsi_schema_location : List[str], optional
            XML schema location, by default None
        xmlns_gpxtpx : str, optional
            ???, by default None
        xmlns_gpxx : str, optional
            ???, by default None
        xmlns_gpxtrk : str, optional
            ???, by default None
        xmlns_wptx1 : str, optional
            ???, by default None
        metadata : Metadata, optional
            Metadata, by default None
        wpt : List[WayPoint], optional
            Way points, by default None
        rte : List[Route], optional
            Routes, by default None
        tracks : List[Track], optional
            List of tracks, by default None
        extensions : Extensions, optional
            Extensions, by default None
        """
        self.tag: str = tag
        self.creator: str = creator
        self.xmlns: str = xmlns
        self.version: str = version

        self.xmlns_xsi: str = xmlns_xsi
        if xsi_schema_location is None:
            self.xsi_schema_location: str = []
        else:
            self.xsi_schema_location: str = xsi_schema_location
        self.xmlns_gpxtpx: str = xmlns_gpxtpx
        self.xmlns_gpxx: str = xmlns_gpxx
        self.xmlns_gpxtrk: str = xmlns_gpxtrk
        self.xmlns_wptx1: str = xmlns_wptx1

        self.metadata: Metadata = metadata
        if wpt is None:
            self.wpt:List[WayPoint] = []
        else:
            self.wpt:List[WayPoint] = wpt
        if rte is None:
            self.rte: List[Route] = []
        else:
            self.rte: List[Route] = rte
        if tracks is None:
            self.tracks: List[Track] = []
        else:
            self.tracks: List[Track] = tracks
        self.extensions: Extensions = extensions

    def check_xml_schema(self, file_path: str) -> bool:
        """
        Check XML schema.

        Parameters
        ----------
        file_path : str
            File path

        Returns
        -------
        bool
            True if the file follows XML schemas
        """
        schema = None

        # GPX
        if file_path.endswith(".gpx"):
            if self.version == "1.1":
                schema = xmlschema.XMLSchema(files("ezgpx.schemas").joinpath("gpx_1_1/gpx.xsd"))
            elif self.version == "1.0":
                schema = xmlschema.XMLSchema(files("ezgpx.schemas").joinpath("gpx_1_0/gpx.xsd"))
            else:
                logging.error("Unable to check XML schema (unsupported GPX version)")
                return False

        # KML
        elif file_path.endswith(".kml"):
            schema = xmlschema.XMLSchema(files("ezgpx.schemas").joinpath("kml_2_2/ogckml22.xsd"))

        # KMZ
        elif file_path.endswith(".kmz"):
            return False

        # FIT
        elif file_path.endswith(".fit"):
            logging.error("Unable to check XML schema (fit files are not XML files)")
            return False
        
        # NOT SUPPORTED
        else:
            logging.error("Unable to check XML schema (unable to identify file type)")
            return False

        if schema is not None:
            return schema.is_valid(file_path)
        else:
            logging.error("Unable to check XML schema (unable to load XML schema)")
            return False
            
    def check_xml_extensions_schemas(self, file_path: str) -> bool:
        gpx_schemas = [s for s in self.xsi_schema_location if s.endswith(".xsd")]
        for gpx_schema in gpx_schemas:
            logging.debug(f"schema = {gpx_schema}")
            schema = xmlschema.XMLSchema(gpx_schema)
            if not schema.is_valid(file_path):
                logging.error(f"File does not follow {gpx_schema}")
                return False

    def name(self) -> str:
        """
        Return activity name.

        Returns
        -------
        str
            Activity name.
        """
        return self.tracks[0].name
    
    def set_name(self, new_name: str) -> None:
        """
        Set name.

        Parameters
        ----------
        new_name : str
            New name.
        """
        self.tracks[0].name = new_name

    def nb_points(self) -> int:
        """
        Compute the number of points in the GPX.

        Returns
        -------
        int
            Number of points in the GPX.
        """
        nb_pts = 0
        for track in self.tracks:
            for track_segment in track.trkseg:
                nb_pts += len(track_segment.trkpt)
        return nb_pts
    
    def first_point(self) -> WayPoint:
        """
        Return GPX first point.

        Returns
        -------
        WayPoint
            First point.
        """
        return self.tracks[0].trkseg[0].trkpt[0]

    def last_point(self) -> WayPoint:
        """
        Return GPX last point.

        Returns
        -------
        WayPoint
            Last point.
        """
        return self.tracks[-1].trkseg[-1].trkpt[-1]
    
    def bounds(self) -> Tuple[float, float, float, float]:
        """
        Find minimum and maximum latitude and longitude.

        Returns
        -------
        Tuple[float, float, float, float]
            Min latitude, min longitude, max latitude, max longitude.
        """
        min_lat = self.tracks[0].trkseg[0].trkpt[0].lat
        min_lon = self.tracks[0].trkseg[0].trkpt[0].lon
        max_lat = min_lat
        max_lon = min_lon

        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.lat < min_lat:
                        min_lat = track_point.lat
                    if track_point.lon < min_lon:
                        min_lon = track_point.lon
                    if track_point.lat > max_lat:
                        max_lat = track_point.lat
                    if track_point.lon > max_lon:
                        max_lon = track_point.lon
        return min_lat, min_lon, max_lat, max_lon


    def center(self) -> Tuple[float, float]:
        """
        Compute the center coordinates of the track.

        Returns
        -------
        Tuple[float, float]
            Latitude and longitude of the center point.
        """
        min_lat, min_lon, max_lat, max_lon = self.bounds()
        center_lat = min_lat + (max_lat - min_lat) / 2
        center_lon = min_lon + (max_lon - min_lon) / 2
        return center_lat, center_lon

    def distance(self) -> float:
        """
        Compute the total distance (meters) of the tracks contained in the Gpx element.

        Returns
        -------
        float
            Distance (meters).
        """
        dst = 0
        previous_point = self.tracks[0].trkseg[0].trkpt[0]
        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    dst += haversine_distance(previous_point, track_point)
                    previous_point = track_point
        return dst
    
    def ascent(self) -> float:
        """
        Compute the total ascent (meters) of the tracks contained in the Gpx element.

        Returns
        -------
        float
            Ascent (meters).
        """
        ascent = 0
        previous_elevation = self.tracks[0].trkseg[0].trkpt[0].ele
        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ele > previous_elevation:
                        ascent += track_point.ele - previous_elevation
                    previous_elevation = track_point.ele
        return ascent
    
    def descent(self) -> float:
        """
        Compute the total descent (meters) of the tracks contained in the Gpx element.

        Returns
        -------
        float
            Descent (meters).
        """
        descent = 0
        previous_elevation = self.tracks[0].trkseg[0].trkpt[0].ele
        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ele < previous_elevation:
                        descent += previous_elevation - track_point.ele
                    previous_elevation = track_point.ele
        return descent
    
    def compute_points_ascent_rate(self) -> None:
        """
        Compute ascent rate at each point.
        """
        previous_point = self.first_point()

        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    distance = haversine_distance(previous_point, track_point)
                    ascent = track_point.ele - previous_point.ele
                    try:
                        track_point.ascent_rate = (ascent * 100) / distance
                        logging.debug(f"distance={distance} | ascent={ascent} | ascent_rate={track_point.ascent_rate}")
                    except:
                        track_point.ascent_rate = 0.0
                    previous_point = track_point

    def min_ascent_rate(self) -> float:
        """
        Return activity minimum ascent rate.

        Returns
        -------
        float
            Minimum ascent rate.
        """
        min_ascent_rate = 100.0
        self.compute_points_ascent_rate()

        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ascent_rate < min_ascent_rate:
                        min_ascent_rate = track_point.ascent_rate

        return min_ascent_rate 

    def max_ascent_rate(self) -> float:
        """
        Return activity maximum ascent rate.

        Returns
        -------
        float
            Maximum ascent rate.
        """
        max_ascent_rate = -1.0
        self.compute_points_ascent_rate()

        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ascent_rate > max_ascent_rate:
                        max_ascent_rate = track_point.ascent_rate

        return max_ascent_rate                
    
    def min_elevation(self) -> float:
        """
        Compute minimum elevation (meters) in the tracks contained in the Gpx element.

        Returns
        -------
        float
            Minimum elevation (meters).
        """
        min_elevation = self.tracks[0].trkseg[0].trkpt[0].ele
        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ele < min_elevation:
                        min_elevation = track_point.ele
        return min_elevation
    
    def max_elevation(self) -> float:
        """
        Compute maximum elevation (meters) in the tracks contained in the Gpx element.

        Returns
        -------
        float
            Maximum elevation (meters).
        """
        max_elevation = self.tracks[0].trkseg[0].trkpt[0].ele
        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ele > max_elevation:
                        max_elevation = track_point.ele
        return max_elevation
    
    def utc_start_time(self) -> datetime:
        """
        Return the activity UTC start time.

        Returns
        -------
        datetime
            UTC start time.
        """
        return self.tracks[0].trkseg[0].trkpt[0].time
    
    def utc_stop_time(self):
        """
        Return the activity UTC stop time.

        Returns
        -------
        datetime
            UTC stop time.
        """
        return self.tracks[-1].trkseg[-1].trkpt[-1].time
    
    def start_time(self) -> datetime:
        """
        Return the activity start time.

        Returns
        -------
        datetime
            Start time.
        """
        start_time = None
        try:
            start_time = self.tracks[0].trkseg[0].trkpt[0].time.replace(tzinfo=timezone.utc).astimezone(tz=None) 
        except:
            logging.error("Unable to find activity start time")
        return start_time
    
    def stop_time(self) -> datetime:
        """
        Return the activity stop time.

        Returns
        -------
        datetime
            Stop time.
        """
        stop_time = None
        try:
            stop_time = self.tracks[-1].trkseg[-1].trkpt[-1].time.replace(tzinfo=timezone.utc).astimezone(tz=None)
        except:
            logging.error("Unable to find activity stop time")
        return stop_time
    
    def total_elapsed_time(self) -> datetime:
        """
        Compute the total elapsed time.

        Returns
        -------
        datetime
            Total elapsed time.
        """
        total_elapsed_time = None
        try:
            total_elapsed_time = self.stop_time() - self.start_time()
        except:
            logging.error("Unable to compute activity total elapsed time")
        return total_elapsed_time
    
    def stopped_time(self, tolerance: float = 2.45) -> datetime:
        """
        Compute the stopped time during activity.

        Parameters
        ----------
        tolerance : float, optional
            Maximal distance between two points for movement, by default 2.45
            (According to my tests with strava_run_1.gpx and the data on Strava)

        Returns
        -------
        datetime
            Stopped time.
        """
        stopped_time = self.start_time() - self.start_time() # Better way to do it?

        previous_point = self.tracks[0].trkseg[0].trkpt[0]

        for track in self.tracks:
            for segment in track.trkseg:
                for point in segment.trkpt:
                    if haversine_distance(previous_point, point) < tolerance:
                        stopped_time += point.time - previous_point.time
                    previous_point = point

        return stopped_time
    
    def moving_time(self) -> datetime:
        """
        Compute the moving time during the activity.

        Returns
        -------
        datetime
            Moving time.
        """
        return self.total_elapsed_time() - self.stopped_time()
    
    def avg_speed(self) -> float:
        """
        Compute the average speed (kilometers per hour) during the activity.

        Returns
        -------
        float
            Average speed (kilometers per hour).
        """
        # Compute and convert total elapsed time
        total_elapsed_time = self.total_elapsed_time()
        total_elapsed_time = total_elapsed_time.total_seconds() / 3600

        # Compute and convert distance
        distance = self.distance() / 1000

        return distance/total_elapsed_time
    
    def avg_moving_speed(self) -> float:
        """
        Compute the average moving speed (kilometers per hour) during the activity.

        Returns
        -------
        float
            Average moving speed (kilometers per hour).
        """
        # Compute and convert moving time
        moving_time = self.moving_time()
        moving_time = moving_time.total_seconds() / 3600

        # Compute and convert distance
        distance = self.distance() / 1000

        return distance / moving_time
    
    def compute_points_speed(self) -> None:
        """
        Compute speed (kilometers per hour) at each track point.
        """
        previous_point = self.first_point()

        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    distance = haversine_distance(previous_point, track_point) / 1000 # Convert to kilometers
                    time = (track_point.time - previous_point.time).total_seconds() / 3600 # Convert to hours
                    try:
                        track_point.speed = distance / time
                    except:
                        track_point.speed = 0.0
                    previous_point = track_point

    def min_speed(self) -> float:
        """
        Return the minimum speed during the activity.

        Returns
        -------
        float
            Minimum speed.
        """
        min_speed = 1000.0
        self.compute_points_speed()

        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.speed < min_speed:
                        min_speed = track_point.speed

        return min_speed

    def max_speed(self) -> float:
        """
        Return the maximum speed during the activity.

        Returns
        -------
        float
            Maximum speed.
        """
        max_speed = -1.0
        self.compute_points_speed()

        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.speed > max_speed:
                        max_speed = track_point.speed

        return max_speed
    
    def avg_pace(self) -> float:
        """
        Compute the average pace (minute per kilometer) during the activity.

        Returns
        -------
        float
            Average pace (minute per kilometer).
        """
        return 60.0 / self.avg_speed()
    
    def avg_moving_pace(self) -> float:
        """
        Compute the average moving pace (minute per kilometer) during the activity.

        Returns
        -------
        float
            Average moving pace (minute per kilometer).
        """
        return 60.0 / self.avg_moving_speed()
    
    def compute_points_pace(self) -> None:
        """
        Compute pace at each track point.
        """
        self.compute_points_speed()

        for track in self.tracks:
            for segment in track.trkseg:
                for point in segment.trkpt:
                    try:
                        point.pace = 60.0 / point.speed
                    except:
                        point.pace = self.avg_moving_pace() # Fill with average moving space (first point)

    def min_pace(self) -> float:
        """
        Return the minimum pace during the activity.

        Returns
        -------
        float
            Minimum pace.
        """
        min_pace = 1000.0
        self.compute_points_pace()

        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.pace < min_pace:
                        min_pace = track_point.pace

        return min_pace

    def max_pace(self) -> float:
        """
        Return the maximum pace during the activity.

        Returns
        -------
        float
            Maximum pace.
        """
        max_pace = -1.0
        self.compute_points_pace()

        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.pace > max_pace:
                        max_pace = track_point.pace

        return max_pace
    
    def compute_points_ascent_speed(self) -> None:
        """
        Compute ascent speed (kilometers per hour) at each track point.
        """
        previous_point = self.first_point()

        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    ascent = (track_point.ele - previous_point.ele) / 1000 # Convert to kilometers
                    time = (track_point.time - previous_point.time).total_seconds() / 3600 # Convert to hours
                    try:
                        track_point.ascent_speed = ascent / time
                    except:
                        track_point.ascent_speed = 0.0
                    previous_point = track_point

    def min_ascent_speed(self) -> float:
        """
        Return the minimum ascent speed (kilometers per hour) during the activity.

        Returns
        -------
        float
            Minimum ascent speed.
        """
        min_ascent_speed = 1000.0
        self.compute_points_ascent_speed()

        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ascent_speed < min_ascent_speed:
                        min_ascent_speed = track_point.ascent_speed

        return min_ascent_speed

    def max_ascent_speed(self) -> float:
        """
        Return the maximum ascent speed (kilometers per hour) during the activity.

        Returns
        -------
        float
            Maximum ascent speed.
        """
        max_ascent_speed = -1.0
        self.compute_points_ascent_speed()

        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ascent_speed > max_ascent_speed:
                        max_ascent_speed = track_point.ascent_speed

        return max_ascent_speed
    
    def remove_points(self, remove_factor: int = 2):
        count = 0
        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if count % remove_factor == 0:
                        track_segment.trkpt.remove(track_point)
                        count += 1

    def remove_gps_errors(self, error_distance=100):
        """
        Remove GPS errors.

        Parameters
        ----------
        error_distance : int, optional
            Error threshold distance (meters) between two points, by default 100

        Returns
        -------
        _type_
            List of removed points (GPS errors).
        """
        previous_point = None
        gps_errors = []

        for track in self.tracks:
            for track_segment in track.trkseg:

                new_trkpt = []

                for track_point in track_segment.trkpt:
                    # GPS error
                    if previous_point is not None and haversine_distance(previous_point,
                                                                         track_point) > error_distance:
                        logging.warning(f"Point {track_point} has been removed (GPS error)")
                        gps_errors.append(track_point)
                    # No GPS error
                    else:
                        new_trkpt.append(track_point)
                        previous_point = track_point

                track_segment.trkpt = new_trkpt

        return gps_errors

    def remove_close_points(self, min_dist: float = 1, max_dist: float = 10):
        """
        Remove points that are to close together.

        Parameters
        ----------
        min_dist : float, optional
            Minimal distance between two points, by default 1
        max_dist : float, optional
            Maximal distance between two points, by default 10
        """
        point_1 = None
        point_2 = None
        
        for track in self.tracks:
            for segment in track.trkseg:

                new_trkpt = []

                for point in segment.trkpt:
                    if point_1 is None:
                        point_1 = point
                        new_trkpt.append(point_1)
                    elif point_2 is None:
                        point_2 = point
                    else:
                        if ((haversine_distance(point_1, point_2) < min_dist
                             or haversine_distance(point_2, point) < min_dist)
                            and haversine_distance(point_1, point) < max_dist):
                            point_2 = point
                        else:
                            new_trkpt.append(point_2)
                            point_1 = point_2
                            point_2 = point

                segment.trkpt = new_trkpt

    def simplify(self, epsilon):
        """
        Simplify GPX tracks using Ramer-Douglas-Peucker algorithm.

        Parameters
        ----------
        epsilon : _type_
            Tolerance.
        """
        for track in self.tracks:
            for segment in track.trkseg:
                segment.trkpt = ramer_douglas_peucker(segment.trkpt, epsilon)

    def to_dataframe(
            self,
            projection: bool = False,
            elevation: bool = True,
            speed: bool = False,
            pace: bool = False,
            ascent_rate: bool = False,
            ascent_speed: bool = False) -> pd.DataFrame:
        """
        Convert GPX object to Pandas Dataframe.

        Parameters
        ----------
        projection : bool, optional
            Toggle projection, by default False
        elevation : bool, optional
            Toggle elevation, by default True
        speed : bool, optional
            Toggle speed, by default False
        pace : bool, optional
            Toggle pace, by default False
        ascent_rate : bool, optional
            Toggle ascent rate, by default False
        ascent_speed : bool, optional
            Toggle ascent speed, by default False

        Returns
        -------
        pd.DataFrame
            Dataframe containing data from GPX.
        """
        test_point = self.first_point()
        if projection and test_point._x is None:
            logging.warning(f"Converting GPX to dataframe with missing projection data.")
        if speed and test_point.speed is None:
            self.compute_points_speed()
        if pace and test_point.pace is None:
            self.compute_points_pace()
        if ascent_rate and test_point.ascent_rate is None:
            self.compute_points_ascent_rate()
        if ascent_speed and test_point.ascent_speed is None:
            self.compute_points_ascent_speed()

        route_info = []
        for track in self.tracks:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    track_point_dict = {
                        "lat": track_point.lat,
                        "lon": track_point.lon
                    }
                    if elevation:
                        if track_point.ele is not None:
                            track_point_dict["ele"] = track_point.ele
                        else:
                            track_point_dict["ele"] = 0
                    if projection:
                        track_point_dict["x"] = track_point._x
                        track_point_dict["y"] = track_point._y
                    if speed:
                        track_point_dict["speed"] = track_point.speed
                    if pace:
                        track_point_dict["pace"] = track_point.pace
                    if ascent_rate:
                        track_point_dict["ascent_rate"] = track_point.ascent_rate
                    if ascent_speed:
                        track_point_dict["ascent_speed"] = track_point.ascent_speed
                    route_info.append(track_point_dict)
        df = pd.DataFrame(route_info)
        return df
    
    def to_csv(
            self,
            path: str = None,
            sep: str = ",",
            columns: List[str] = None,
            header: bool = True,
            index: bool = False) -> Union[str, None]:
        """
        Write the GPX object track coordinates to a .csv file.

        Parameters
        ----------
        path : str, optional
            Path to the .csv file, by default None
        sep : str, optional
            Separator, by default ","
        columns : List[str], optional
            List of columns to write, by default None
        header : bool, optional
            Toggle header, by default True
        index : bool, optional
             Toggle index, by default False

        Returns
        -------
        str
            CSV like string if path is set to None.
        """
        if columns is None:
            columns =  ["lat", "lon"]

        elevation = False
        projection = False
        speed = False
        pace = False
        ascent_rate = False
        ascent_speed = False
        
        if "ele" in columns:
            elevation = True
        if "x" in columns or "y" in columns:
            projection = True
        if "speed" in columns:
            speed = True
        if "pace" in columns:
            pace = True
        if "ascent_rate" in columns:
            ascent_rate = True
        if "ascent_speed" in columns:
            ascent_speed = True

        return self.to_dataframe(projection, elevation, speed, pace, ascent_rate, ascent_speed).to_csv(path, sep=sep, columns=columns, header=header, index=index)

    def project(self, projection: str):
        """
        Project tracks.

        Parameters
        ----------
        projection : str
            Projection.
        """
        for track in self.tracks:
            track.project(projection)


from typing import Tuple
from math import pi, radians, floor, log, cos, sin, tan
from ..utils import EARTH_RADIUS


def web_mercator_projection(latitude: float, longitude: float, zoom_level: int = 1) -> Tuple[int, int]:
    """
    https://en.wikipedia.org/wiki/Web_Mercator_projection

    Parameters
    ----------
    latitude : float
        Point latitude.
    longitude : float
        Point longitude.
    zoom_level : int, optional
        Zoom level, by default 1

    Returns
    -------
    Tuple[int, int]
        Point coordinates in the web-mercator projection.
    """
    # WARNING -> Poles: see link above

    # Convert latitude and longitude to radians
    latitude, longitude = radians(latitude), radians(longitude)

    # Compute coordinates
    const = (2**zoom_level)/(2 * pi)
    x = const * (longitude + pi)
    y = const * (pi - log(tan(pi/4 + latitude/2)))

    # return floor(x), floor(y)
    return x, y  # ???


def lambert_conformal_conic_projection(latitude: float, longitude: float, ref_latitude: float, ref_longitude: float, standard_parallel_1: float, standard_parallel_2: float) -> Tuple[float, float]:
    """
    https://en.wikipedia.org/wiki/Lambert_conformal_conic_projection

    Parameters
    ----------
    latitude : float
        Point latitude.
    longitude : float
        Point longitude.
    ref_latitude : float
        Reference point latitude.
    ref_longitude : float
        Reference point longitude.
    standard_parallel_1 : float
        First standard parallel.
    standard_parallel_2 : float
        Second standard parallel.

    Returns
    -------
    Tuple[float, float]
        Point coordinates in the Lambert conformal conic projection.
    """
    latitude = radians(latitude)
    longitude = radians(longitude)
    ref_latitude = radians(ref_latitude)
    ref_longitude = radians(ref_longitude)
    standard_parallel_1 = radians(standard_parallel_1)
    standard_parallel_2 = radians(standard_parallel_2)

    if standard_parallel_1 != standard_parallel_2:
        n = log(cos(standard_parallel_1) * (1/cos(standard_parallel_2))) / log(tan(0.75 *
                                                                                   pi + 0.5*standard_parallel_2) * (1/tan(0.75*pi + 0.5*standard_parallel_1)))
    else:
        n = sin(standard_parallel_1)

    F = cos(standard_parallel_1) * \
        (tan(0.75*pi + 0.5*standard_parallel_1))**n / n
    rho = EARTH_RADIUS*F*(1/tan(0.75*pi + 0.5*latitude))**n
    rho_0 = EARTH_RADIUS*F*(1/tan(0.75*pi + 0.5*ref_latitude))**n

    x = rho*sin(n*(longitude - ref_longitude))
    y = rho_0 - rho*cos(n*(longitude - ref_longitude))
    
    print(f"n={n}")
    print(f"F={F}")
    print(f"rho={rho}")
    print(f"rho_0={rho_0}")
    print(f"x,y={x},{y}")

    return x, y


import os
from typing import Optional, Union, List, Tuple, Dict
import logging
import xml.etree.ElementTree as ET
from datetime import datetime

from ..writer import Writer
from ..gpx_elements import Bounds, Copyright, Email, Extensions, Gpx, Link, Metadata, Person, PointSegment, Point, Route, TrackSegment, Track, WayPoint
from ..gpx_parser import GPXParser

class GPXWriter(Writer):
    """
    GPX file writer.
    """

    def __init__(
            self,
            gpx: Gpx = None,
            file_path: str = None,
            properties: bool = True,
            metadata: bool = True,
            way_points: bool = True,
            routes: bool = True,
            extensions: bool = True,
            ele: bool = True,
            time: bool = True,
            precisions: Dict = None,
            time_format: str = None) -> None:
        """
        Initialize GPXWriter instance.

        Args:
            gpx (Gpx, optional): Gpx instance to write. Defaults to None.
            file_path (str, optional): Path to the file to write. Defaults to None.
            properties (bool, optional): Toggle properties writting. Defaults to True.
            metadata (bool, optional): Toggle metadata writting. Defaults to True.
            way_point (bool, optional): Toggle way points writting. Defaults to True.
            routes (bool, optional): Toggle routes writting. Defaults to True.
            extensions (bool, optional): Toggle extensions writting. Defaults to True.
            ele (bool, optional): Toggle elevation writting. Defaults to True.
            time (bool, optional): Toggle time writting. Defaults to True.
            precisions (dict, optional): Decimal precision for each type of value. Defaults to None.
            time_format (dict, optional): Time format. Defaults to None.
        """
        super().__init__(gpx, file_path)
        self.gpx_string: str = ""

        # Parameters
        self.properties: bool = properties
        self.metadata: bool = metadata
        self.way_points: bool = way_points
        self.routes: bool = routes
        self.extensions: bool = extensions
        self.ele: bool = ele
        self.time: bool = time

        self.precisions: Dict = precisions
        self.time_format = time_format

        self.gpx_root = None
    
    def add_bounds(self, element: ET.Element, bounds: Bounds) -> ET.Element:
        """
        Add Bounds instance to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            bounds (Bounds): Bounds instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        if bounds is not None:
            bounds_ = ET.SubElement(element, bounds.tag)
            bounds_, _ = self.add_subelement_number(bounds_, "minlat", bounds.minlat, self.precisions["lat_lon"])
            bounds_, _ = self.add_subelement_number(bounds_, "minlon", bounds.minlon, self.precisions["lat_lon"])
            bounds_, _ = self.add_subelement_number(bounds_, "maxlat", bounds.maxlat, self.precisions["lat_lon"])
            bounds_, _ = self.add_subelement_number(bounds_, "maxlon", bounds.maxlon, self.precisions["lat_lon"])
        return element
    
    def add_copyright(self, element: ET.Element, copyright: Copyright) -> ET.Element:
        """
        Add Copyright instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            copyright (Copyright): Copyright instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        if copyright is not None:
            copyright_ = ET.SubElement(element, copyright.tag)
            if copyright.author is not None:
                self.setIfNotNone(copyright_, "author", copyright.author)
            copyright_, _ = self.add_subelement(copyright_, "year", str(copyright.year))
            copyright_, _ = self.add_subelement(copyright_, "licence", str(copyright.licence))
        return element
    
    def add_email(self, element: ET.Element, email: Email) -> ET.Element:
        """
        Add Email instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            email (Email): Email instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        if email is not None:
            email_ = ET.SubElement(element, email.tag)
            if email.id is not None:
                self.setIfNotNone(email_, "id", email.id)
            if email.domain is not None:
                self.setIfNotNone(email_, "domain", email.domain)
        return element

    def add_extensions(self, element: ET.Element, extensions: Extensions) -> ET.Element:
        """
        Add Extensions instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            extensions (Extensions): Extensions instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        if extensions is not None:
            extensions_ = ET.SubElement(element, extensions.tag)
            extensions_, _ = self.add_subelement(extensions_, "display_color", extensions.display_color)
            extensions_, _ = self.add_subelement_number(extensions_, "distance", extensions.distance, self.precisions["distance"])
            extensions_, _ = self.add_subelement_number(extensions_, "total_elapsed_time", extensions.total_elapsed_time, self.precisions["duration"])
            extensions_, _ = self.add_subelement_number(extensions_, "moving_time", extensions.moving_time, self.precisions["duration"])
            extensions_, _ = self.add_subelement_number(extensions_, "stopped_time", extensions.stopped_time, self.precisions["duration"])
            extensions_, _ = self.add_subelement_number(extensions_, "moving_speed", extensions.moving_speed, self.precisions["speed"])
            extensions_, _ = self.add_subelement_number(extensions_, "max_speed", extensions.max_speed, self.precisions["speed"])
            extensions_, _ = self.add_subelement_number(extensions_, "max_elevation", extensions.max_elevation, self.precisions["elevation"])
            extensions_, _ = self.add_subelement_number(extensions_, "min_elevation", extensions.min_elevation, self.precisions["elevation"])
            extensions_, _ = self.add_subelement_number(extensions_, "ascent", extensions.ascent, self.precisions["elevation"])
            extensions_, _ = self.add_subelement_number(extensions_, "descent", extensions.descent, self.precisions["elevation"])
            extensions_, _ = self.add_subelement_number(extensions_, "avg_ascent_rate", extensions.avg_ascent_rate, self.precisions["rate"])
            extensions_, _ = self.add_subelement_number(extensions_, "max_ascent_rate", extensions.max_descent_rate, self.precisions["rate"])
            extensions_, _ = self.add_subelement_number(extensions_, "avg_descent_rate", extensions.avg_descent_rate, self.precisions["rate"])
            extensions_, _ = self.add_subelement_number(extensions_, "max_descent_rate", extensions.max_descent_rate, self.precisions["rate"])
        return element
    
    def add_link(self, element: ET.Element, link: Link) -> ET.Element:
        """
        Add Link instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            link (Link): Link instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        if link is not None:
            link_ = ET.SubElement(element, link.tag)
            if link.href is not None:
                self.setIfNotNone(link_, "href", link.href)
            link_, _ = self.add_subelement(link_, "text", link.text)
            link_, _ = self.add_subelement(link_, "type", link.type)
        return element

    def add_metadata(self, element: ET.Element, metadata: Metadata) -> ET.Element:
        """
        Add Metadata instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            metadata (Metadata): Metadata instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        if metadata is not None:
            metadata_ = ET.SubElement(element, metadata.tag)
            metadata_, _ = self.add_subelement(metadata_, "name", metadata.name)
            metadata_, _ = self.add_subelement(metadata_, "desc", metadata.desc)
            metadata_ = self.add_person(metadata_, metadata.author)
            metadata_ = self.add_copyright(metadata_, metadata.copyright)
            metadata_ = self.add_link(metadata_, metadata.link)
            metadata_, _ = self.add_subelement_time(metadata_, "time", metadata.time, self.time_format)
            metadata_, _ = self.add_subelement(metadata_, "keywords", metadata.keywords)
            metadata_ = self.add_bounds(metadata_, metadata.bounds)
            metadata_ = self.add_extensions(metadata_, metadata.extensions)
        return element
    
    def add_person(self, element: ET.Element, person: Person) -> ET.Element:
        """
        Add Person instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            person (Person): Person instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        if person is not None:
            person_ = ET.SubElement(element, person.tag)
            person_, _ = self.add_subelement(person_, "name", person.name)
            person_ = self.add_email(person_, person.email)
            person_ = self.add_link(person_, person.link)
        return element
    
    def add_point_segment(self, element: ET.Element, point_segment: PointSegment) -> ET.Element:
        """
        Add PointSegment instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            point_segment (PointSegment): PointSegment instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        if point_segment is not None:
            pass
        return element

    def add_point(self, element: ET.Element, point: Point) -> ET.Element:
        """
        Add Point instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            point (Point): Point instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        if point is not None:
            point_ = ET.SubElement(element, point.tag)
            self.setIfNotNone(point_, "lat", "{:.{}f}".format(point.lat, self.precisions["lat_lon"]))
            self.setIfNotNone(point_, "lon", "{:.{}f}".format(point.lon, self.precisions["lat_lon"]))
            point_ = self.add_subelement_number(point_, "ele", point.ele, self.precisions["elevation"])
            point_, _ = self.add_subelement_time(point_, "time", point.time, self.time_format)
        return element

    def add_route(self, element: ET.Element, route: Route) -> ET.Element:
        """
        Add Route instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            route (Route): Route instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        if route is not None:
            route_ = ET.SubElement(element, route.tag)
            route_, _ = self.add_subelement(route_, "name", route.name)
            route_, _ = self.add_subelement(route_, "cmt", route.cmt)
            route_, _ = self.add_subelement(route_, "desc", route.desc)
            route_, _ = self.add_subelement(route_, "src", route.src)
            route_ = self.add_link(route_, route.link)
            route_, _ = self.add_subelement_number(route_, "number", route.src, 0)
            route_, _ = self.add_subelement(route_, "type", route.type)
            route_ = self.add_extensions(route_, route.extensions)
            for way_point in route.rtept:
                route_ = self.add_way_point(route_, way_point)
        return element

    def add_track_segment(self, element: ET.Element, track_segment: TrackSegment) -> ET.Element:
        """
        Add TrackSegment instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            track_segment (TrackSegment): TrackSegment instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        if track_segment is not None:
            track_segment_ = ET.SubElement(element, track_segment.tag)
            for track_point in track_segment.trkpt:
                track_segment_ = self.add_way_point(track_segment_, track_point)
        return element

    def add_track(self, element: ET.Element, track: Track) -> ET.Element:
        """
        Add Track instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            track (Track): Track instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        if track is not None:
            track_ = ET.SubElement(element, track.tag)
            track_, _ = self.add_subelement(track_, "name", track.name)
            for track_segment in track.trkseg:
                track_ = self.add_track_segment(track_, track_segment)
        return element
    
    def add_way_point(self, element: ET.Element, way_point: WayPoint) -> ET.Element:
        """
        Add WayPoint instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            way_point (WayPoint): WayPoint instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        if way_point is not None:
            way_point_ = ET.SubElement(element, way_point.tag)
            self.setIfNotNone(way_point_, "lat", "{:.{}f}".format(way_point.lat, self.precisions["lat_lon"]))
            self.setIfNotNone(way_point_, "lon", "{:.{}f}".format(way_point.lon, self.precisions["lat_lon"]))
            way_point_, _ = self.add_subelement_number(way_point_, "ele", way_point.ele, self.precisions["elevation"])
            way_point_, _ = self.add_subelement_time(way_point_, "time", way_point.time, self.time_format)
            way_point_, _ = self.add_subelement_number(way_point_, "magvar", way_point.mag_var, self.precisions["default"])
            way_point_, _ = self.add_subelement_number(way_point_, "geoidheight", way_point.geo_id_height, self.precisions["default"])
            way_point_, _ = self.add_subelement(way_point_, "name", way_point.name)
            way_point_, _ = self.add_subelement(way_point_, "cmt", way_point.cmt)
            way_point_, _ = self.add_subelement(way_point_, "desc", way_point.desc)
            way_point_, _ = self.add_subelement(way_point_, "src", way_point.src)
            way_point_ = self.add_link(way_point_, way_point.link)
            way_point_, _ = self.add_subelement(way_point_, "src", way_point.src)
            way_point_, _ = self.add_subelement(way_point_, "type", way_point.type)
            way_point_, _ = self.add_subelement(way_point_, "fix", way_point.fix)
            way_point_, _ = self.add_subelement_number(way_point_, "sat", way_point.sat, 0)
            way_point_, _ = self.add_subelement_number(way_point_, "hdop", way_point.hdop, self.precisions["default"])
            way_point_, _ = self.add_subelement_number(way_point_, "vdop", way_point.vdop, self.precisions["default"])
            way_point_, _ = self.add_subelement_number(way_point_, "pdop", way_point.pdop, self.precisions["default"])
            way_point_, _ = self.add_subelement_number(way_point_, "ageofgpsdata", way_point.age_of_gps_data, self.precisions["default"])
            way_point_, _ = self.add_subelement_number(way_point_, "dgpsid", way_point.dgpsid, 0)
            way_point_ = self.add_extensions(way_point_, way_point.extensions)
        return element

    def createSchemaLocationString(self, xsi_schema_location: list[str]):
        """
        Create schema location string to write in GPX file.

        Parameters
        ----------
        xsi_schema_location : list[str]
            List of schema locations.

        Returns
        -------
        str
            Schema location string to write in GPX file.
        """
        schema_location_string = ""
        for loc in xsi_schema_location:
            schema_location_string += loc
            schema_location_string += " "
        schema_location_string = schema_location_string[:-1]
        return schema_location_string
         
    def add_properties_garmin(self) -> None:
        """
        Add Garmin style properties to the GPX root element.
        """
        self.setIfNotNone(self.gpx_root, "xmlns", self.gpx.xmlns)
        self.setIfNotNone(self.gpx_root, "creator", self.gpx.creator)
        self.setIfNotNone(self.gpx_root, "version", self.gpx.version)
        schema_location_string = self.createSchemaLocationString(self.gpx.xsi_schema_location)
        self.setIfNotNone(self.gpx_root, "xsi:schemaLocation", schema_location_string)
        self.setIfNotNone(self.gpx_root, "xmlns:xsi", self.gpx.xmlns_xsi)
    
    def add_properties_strava(self) -> None:
        """
        Add Strava style properties to the GPX root element.
        """
        self.setIfNotNone(self.gpx_root, "creator", self.gpx.creator)
        self.setIfNotNone(self.gpx_root, "xmlns:xsi", self.gpx.xmlns_xsi)
        schema_location_string = self.createSchemaLocationString(self.gpx.xsi_schema_location)
        self.setIfNotNone(self.gpx_root, "xsi:schemaLocation", schema_location_string)
        self.setIfNotNone(self.gpx_root, "version", self.gpx.version)
        self.setIfNotNone(self.gpx_root, "xmlns", self.gpx.xmlns)
    
    def add_root_properties(self) -> None:
        """
        Add properties to the GPX root element.
        """
        logging.info("Preparing properties...")

        if self.gpx.creator in ["eTrex 32x"]:
             self.add_properties_garmin()
        elif self.gpx.creator in ["StravaGPX"]:
             self.add_properties_strava()
        else:
             self.add_properties_garmin() # Default to Garmin style

    def add_root_metadata(self) -> None:
        """
        Add metadata element to the GPX root element.
        """
        logging.info("Preparing metadata...")

        self.gpx_root = self.add_metadata(self.gpx_root, self.gpx.metadata)
    
    def add_root_way_points(self) -> None:
        """
        Add wpt elements to the GPX root element.
        """
        logging.info("Preparing way points...")

        for way_point in self.gpx.wpt:
            self.gpx_root = self.add_way_point(self.gpx_root, way_point)
    
    def add_root_routes(self) -> None:
        """
        Add rte elements to the GPX root element.
        """
        logging.info("Preparing routes...")

        for route in self.gpx.rte:
            self.gpx_root = self.add_route(self.gpx_root, route)
    
    def add_root_tracks(self) -> None:
        """
        Add trck elements to the GPX root element.
        """
        logging.info("Preparing tracks...")

        for track in self.gpx.tracks:
            self.gpx_root = self.add_track(self.gpx_root, track)
    
    def add_root_extensions(self) -> None:
        """
        Add extensions element to the GPX root element.
        """
        logging.info("Preparing extensions...")

        if self.gpx.extensions is not None:
            self.gpx_root = self.add_extensions(self.gpx_root, self.gpx.extensions)

    def gpx_to_string(self) -> str:
        """
        Convert Gpx instance to a string (the content of a .gpx file).

        Returns:
            str: String corresponding to the Gpx instance.
        """
        if self.gpx is not None:
            logging.info("Start convertion from GPX to string")
            # Reset string
            self.gpx_string = ""

            # Root
            self.gpx_root = ET.Element("gpx")

            # Properties
            if self.properties:
                self.add_root_properties()

            # Metadata
            if self.metadata:
                self.add_root_metadata()

            # Way points
            if self.way_points:
                self.add_root_way_points()

            # Routes
            if self.routes:
                self.add_root_routes()

            # Tracks
            self.add_root_tracks()

            # Extensions
            if self.extensions:
                self.add_root_extensions()

            # Convert data to string
            logging.info("Converting GPX to string...")
            self.gpx_string = ET.tostring(self.gpx_root, encoding="unicode")
            # self.gpx_string = ET.tostring(gpx_root, encoding="utf-8")

            logging.info(f"GPX successfully converted to string:\n{self.gpx_string}")

            return self.gpx_string

    def write_gpx(self):
        """
        Convert Gpx instance to string and write to file.
        """
        # Open/create GPX file
        try:
            f = open(self.file_path, "w")
        except OSError:
            logging.exception(f"Could not open/read file: {self.file_path}")
            raise
        # Write GPX file
        with f:
            f.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>")
            f.write(self.gpx_string)

    def write(
            self,
            file_path: str,
            xml_schema: bool = False,
            xml_extensions_schemas: bool = False) -> bool:
        """
        Handle writing.

        Args:
            path (str): Path to write the GPX file.
            check_xml_schemas (bool, optional): Toggle schema verification after writting. Defaults to False.
            extensions_schemas (bool, optional): Toggle extensions schema verificaton after writing. Requires internet connection and is not guaranted to work. Defaults to False.

        Returns:
            bool: Return False if written file does not follow checked schemas. Return True otherwise.
        """
        directory_path = os.path.dirname(os.path.realpath(file_path))
        if not os.path.exists(directory_path):
            logging.error("Provided path does not exist")
            return False
        self.file_path = file_path
        self.file_name = os.path.basename(self.file_path)

        # Write .gpx file
        self.gpx_to_string()
        self.write_gpx()

        # Check XML schemas
        res = True
        if xml_schema or xml_extensions_schemas:
            res = self.check_xml_schemas(xml_schema, xml_extensions_schemas)
            
        return res


import logging
import datetime

from .extensions import Extensions
from .link import Link
from ..utils import web_mercator_projection, lambert_conformal_conic_projection

class WayPoint():
    """
    wptType element in GPX file.
    """

    def __init__(
            self,
            tag: str = "wpt",
            lat: float = None,
            lon: float = None,
            ele: float = None,
            time: datetime = None,
            mag_var: float = None, # 0 <= value < 360
            geo_id_height: float = None,
            name: str = None,
            cmt: str = None,
            desc: str = None,
            src: str = None,
            link: Link = None,
            sym: str = None,
            type: str = None,
            fix: str = None, # none, 2d, 3d, dgps, pps
            sat: int = None, # non negative integer
            hdop: float = None,
            vdop: float = None,
            pdop: float = None,
            age_of_gps_data: float = None,
            dgpsid: int = None, # 0 <= value <= 1023
            extensions: Extensions = None) -> None:
        """
        Initialize WayPoint instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "wpt"
        lat : float, optional
            Latitude (degrees), by default None
        lon : float, optional
            Longitude (degrees), by default None
        ele : float, optional
            Elevation (meters), by default None
        time : datetime, optional
            Time, by default None
        mag_var : float, optional
            Magnetic variation (degrees) at the point, by default None
        geo_id_height : float, optional
            Height (meters) of geoid (mean sea level) above WGS84 earth
            ellipsoid. As defined in NMEA GGA message., by default None
        name : str, optional
            Name, by default None
        cmt : str, optional
            Comment, by default None
        desc : str, optional
            Description, by default None
        src : str, optional
            Source, by default None
        link : Link, optional
            Link, by default None
        sym : str, optional
            Text of GPS symbol name. For interchange with other programs,
            use the exact spelling of the symbol as displayed on the GPS.
            If the GPS abbreviates words, spell them out, by default None
        type : str, optional
            Type, by default None
        fix : str, optional
            Type of GPX fix, by default None
        sat : int, optional
            Number of satellites used to calculate the GPX fix, by default None
        hdop : float, optional
            Horizontal dilution of precision, by default None
        vdop : float, optional
            Vertical dilution of precision, by default None
        pdop : float, optional
            Position dilution of precision, by default None
        age_of_gps_data : float, optional
            Number of seconds since last DGPS update, by default None
        dgpsid : int, optional
            ID of DGPS station used in differential correction, by default None
        extensions: Extensions, optional
            Extensions, by default None
        """
        self.tag: str = tag
        self.lat: float = lat
        self.lon: float = lon
        self.ele: float = ele
        self.time: datetime = time
        self.mag_var: float = mag_var
        self.geo_id_height: float = geo_id_height
        self.name: str = name
        self.cmt: str = cmt
        self.desc: str = desc
        self.src: str = src
        self.link: Link = link
        self.sym: str = sym
        self.type: str = type
        self.fix: str = fix
        self.sat: int = sat
        self.hdop: float = hdop
        self.vdop: float = vdop
        self.pdop: float = pdop
        self.age_of_gps_data: float = age_of_gps_data
        self.dgpsid: int = dgpsid
        self.extensions: Extensions = extensions

        # Statistics (for map plotting: https://support.strava.com/hc/en-us/articles/360049869011-Personalized-Stat-Maps)
        self.speed: float = None
        self.pace: float = None
        self.ascent_rate: float = None
        self.ascent_speed: float = None

        # Projection
        self._x: int = None
        self._y: int = None

    def project(self, projection: str):
        """
        Project point.

        Parameters
        ----------
        projection : str
            Projection.
        """
        if projection in ["web_mercator_projection", "web_mercator", "wm"]:
            self._x, self._y = web_mercator_projection(self.lat, self.lon)
        elif projection in ["lambert_conformal_conic_projection", "lambert_conformal_conic", "lcc"]:
            self._x, self._y = lambert_conformal_conic_projection(self.lat, self.lon, 90.0, 135.0, 45.0, 45.0)
        else:
            logging.error(f"Invalid projection: {projection}")



from datetime import datetime

from .link import Link
from .person import Person

class Metadata():
    """
    metadataType element in GPX file.
    """

    def __init__(
            self,
            tag: str = "metadata",
            name: str = None,
            desc: str = None,
            author: Person = None,
            copyright: str = None,
            link: Link = None,
            time: datetime = None,
            keywords: str = None,
            bounds: str = None,
            extensions: str = None) -> None:
        """
        Initialize Metadata instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "metadata"
        name : str, optional
            Name, by default None
        desc : str, optional
            Description, by default None
        author : Person, optional
            Author, by default None
        copyright : str, optional
            Copyright, by default None
        link : Link, optional
            Link, by default None
        time : datetime, optional
            Time, by default None
        keywords : str, optional
            Keywords, by default None
        bounds : str, optional
            Bounds, by default None
        extensions : str, optional
            Extensions, by default None
        """
        self.tag: str = tag
        self.name: str = name
        self.desc: str = desc
        self.author: Person = author
        self.copyright: str = copyright
        self.link: Link = link
        self.time: datetime = time
        self.keywords: str = keywords
        self.bounds: str = bounds
        self.extensions: str = extensions


import os
from typing import Optional, Union, List, Tuple, Dict, NewType
import logging
from zipfile import ZipFile
import webbrowser
from datetime import datetime
import pandas as pd
from math import degrees

from fitparse import FitFile

from matplotlib.axes import Axes
from matplotlib.figure import Figure
import matplotlib.colors
import matplotlib.pyplot as plt
from mpl_toolkits.basemap import Basemap

import gmplot

import folium
from folium.plugins import MiniMap
from folium.features import DivIcon

from ..gpx_elements import Gpx, WayPoint
from ..gpx_parser import GPXParser
from ..kml_parser import KMLParser
from ..fit_parser import FitParser
from ..gpx_writer import GPXWriter
from ..kml_writer import KMLWriter
from ..utils import EARTH_RADIUS

GPX = NewType("GPX", object) # GPX forward declaration for type hint

class GPX():
    """
    High level GPX object.
    """

    def __init__(
            self,
            file_path: Optional[str] = None,
            xml_schema: bool = True,
            xml_extensions_schemas: bool = False) -> None:
        """
        Initialise GPX instance.

        Parameters
        ----------
        file_path : Optional[str], optional
            Path to the file to parse, by default None
        xml_schema : bool, optional
            Toggle schema verification during parsing, by default True
        extensions_schemas : bool, optional
            Toggle extensions schema verificaton during parsing, by default False
        """
        if file_path is not None and os.path.exists(file_path):
            self.file_path: str = file_path
            self.gpx: Gpx = None
            self.gpx_parser: GPXParser = None
            self.kml_parser: KMLParser = None
            self.fit_parser: FitParser = None
            self.precisions: Dict = None
            self.time_format: str = None

            # GPX
            if file_path.endswith(".gpx"):
                self.gpx_parser = GPXParser(file_path, xml_schema, xml_extensions_schemas)
                self.gpx = self.gpx_parser.gpx
                self.precisions = self.gpx_parser.precisions
                self.time_format = self.gpx_parser.time_format

            # KML
            elif file_path.endswith(".kml"):
                self.kml_parser = KMLParser(file_path, xml_schema, xml_extensions_schemas)
                self.gpx = self.kml_parser.gpx
                self.precisions = self.kml_parser.precisions
                self.time_format = self.kml_parser.time_format

            # KMZ
            elif file_path.endswith(".kmz"):
                kmz = ZipFile(file_path, 'r')
                kmls = [info.filename for info in kmz.infolist() if info.filename.endswith(".kml")]
                if "doc.kml" not in kmls:
                    logging.warning("Unable to parse this file: Expected to find doc.kml inside KMZ file.")
                kml = kmz.open("doc.kml", 'r').read()
                self.write_tmp_kml("tmp.kml", kml)
                self.kml_parser = KMLParser("tmp.kml", xml_schema, xml_extensions_schemas)
                self.gpx = self.kml_parser.gpx
                self.precisions = self.kml_parser.precisions
                self.time_format = self.kml_parser.time_format
                os.remove("tmp.kml")
            
            # FIT
            elif file_path.endswith(".fit"):
                self.fit_parser = FitParser(file_path)
                self.gpx = self.fit_parser.gpx
                self.precisions = self.fit_parser.precisions
                self.time_format = self.fit_parser.time_format

            # NOT SUPPORTED
            else:
                logging.error("Unable to parse this type of file...\nYou may consider renaming your file with the proper file extension.")
            self.gpx_writer: GPXWriter = GPXWriter(
                self.gpx, precisions=self.precisions, time_format=self.time_format)
            self.kml_writer: KMLWriter = KMLWriter(
                self.gpx, precisions=self.precisions, time_format=self.time_format)
        else:
            logging.warning("File path does not exist")
            pass

    def __str__(self) -> str:
        return f"file_path = {self.file_path}\ngpx = {self.gpx}"
    
    def check_xml_schema(self) -> bool:
        """
        Check XML schema.

        Returns
        -------
        bool
            True if the file follows XML schemas.
        """
        return self.gpx.check_xml_schema(self.file_path)
    
    def check_xml_extensions_schemas(self) -> bool:
        """
        Check XML extension schemas.

        Returns
        -------
        bool
            True if the file follows XML schemas.
        """
        return self.gpx.check_xml_extensions_schemas(self.file_path)
        
    def file_name(self) -> Union[str, None]:
        """
        Return .gpx file name.

        Returns
        -------
        Union[str, None]
            File name.
        """
        return os.path.basename(self.file_path)

    def name(self) -> str:
        """
        Return activity name.

        Returns
        -------
        str
            Activity name.
        """
        return self.gpx.name()
    
    def set_name(self, new_name: str) -> None:
        """
        Set name.

        Parameters
        ----------
        new_name : str
            New name.
        """
        self.gpx.set_name(new_name)

    def nb_points(self) -> int:
        """
        Return the number of points in the GPX.

        Returns
        -------
        int
            Number of points in the GPX.
        """
        return self.gpx.nb_points()

    def first_point(self) -> WayPoint:
        """
        Return GPX first point.

        Returns
        -------
        WayPoint
            First point.
        """
        return self.gpx.first_point()

    def last_point(self) -> WayPoint:
        """
        Return GPX last point.

        Returns
        -------
        WayPoint
            Last point.
        """
        return self.gpx.last_point()

    def bounds(self) -> Tuple[float, float, float, float]:
        """
        Find minimum and maximum latitude and longitude.

        Returns
        -------
        Tuple[float, float, float, float]
            Min latitude, min longitude, max latitude, max longitude.
        """
        return self.gpx.bounds()

    def center(self) -> Tuple[float, float]:
        """
        Return the coordinates of the center point.

        Returns
        -------
        Tuple[float, float]
            Latitude and longitude of the center point.
        """
        return self.gpx.center()

    def distance(self) -> float:
        """
        Returns the distance (meters) of the tracks contained in the GPX.

        Returns
        -------
        float
            Distance (meters).
        """
        return self.gpx.distance()

    def ascent(self) -> float:
        """
        Returns the ascent (meters) of the tracks contained in the GPX.

        Returns
        -------
        float
            Ascent (meters).
        """
        return self.gpx.ascent()

    def descent(self) -> float:
        """
        Returns the descent (meters) of the tracks contained in the GPX.

        Returns
        -------
        float
            Descent (meters).
        """
        return self.gpx.descent()
    
    def compute_points_ascent_rate(self) -> None:
        """
        Compute ascent rate at each point.
        """
        self.gpx.compute_points_ascent_rate()

    def min_ascent_rate(self) -> float:
        """
        Return activity minimum ascent rate.

        Returns
        -------
        float
            Minimum ascent rate.
        """
        return self.gpx.min_ascent_rate()
    
    def max_ascent_rate(self) -> float:
        """
        Return activity maximum ascent rate.

        Returns
        -------
        float
            Maximum ascent rate.
        """
        return self.gpx.max_ascent_rate()

    def min_elevation(self) -> float:
        """
        Returns the minimum elevation (meters) in the tracks contained in the GPX.

        Returns
        -------
        float
            Minimum elevation (meters).
        """
        return self.gpx.min_elevation()

    def max_elevation(self) -> float:
        """
        Returns the maximum elevation (meters) in the tracks contained in the GPX.

        Returns
        -------
        float
            Maximum elevation (meters).
        """
        return self.gpx.max_elevation()

    def start_time(self) -> datetime:
        """
        Return the activity start time.

        Returns
        -------
        datetime
            Start time.
        """
        return self.gpx.start_time()

    def stop_time(self) -> datetime:
        """
        Return the activity stop time.

        Returns
        -------
        datetime
            Stop time.
        """
        return self.gpx.stop_time()

    def total_elapsed_time(self) -> datetime:
        """
        Return the total elapsed time during the activity.

        Returns
        -------
        datetime
            Total elapsed time.
        """
        return self.gpx.total_elapsed_time()

    def stopped_time(self) -> datetime:
        """
        Return the stopped time during the activity.

        Returns
        -------
        datetime
            Stopped time.
        """
        return self.gpx.stopped_time()

    def moving_time(self) -> datetime:
        """
        Return the moving time during the activity.

        Returns
        -------
        datetime
            Moving time.
        """
        return self.gpx.moving_time()

    def avg_speed(self) -> float:
        """
        Return average speed (kilometers per hour) during the activity.

        Returns
        -------
        float
            Average speed (kilometers per hour).
        """
        return self.gpx.avg_speed()

    def avg_moving_speed(self) -> float:
        """
        Return average moving speed (kilometers per hour) during the activity.

        Returns
        -------
        float
            Average moving speed (kilometers per hour).
        """
        return self.gpx.avg_moving_speed()
    
    def compute_points_speed(self) -> None:
        """
        Compute speed (kilometers per hour) at each track point.
        """
        self.gpx.compute_points_speed()

    def min_speed(self) -> float:
        """
        Return the minimum speed during the activity.

        Returns
        -------
        float
            Minimum speed.
        """
        return self.gpx.min_speed()

    def max_speed(self) -> float:
        """
        Return the maximum speed during the activity.

        Returns
        -------
        float
            Maximum speed.
        """
        return self.gpx.max_speed()
    
    def avg_pace(self) -> float:
        """
        Return average pace (minutes per kilometer) during the activity.

        Returns
        -------
        float
            Average pace (minutes per kilometer).
        """
        return self.gpx.avg_pace()

    def avg_moving_pace(self) -> float:
        """
        Return average moving pace (minutes per kilometer) during the activity.

        Returns
        -------
        float
            Average moving pace (minutes per kilometer).
        """
        return self.gpx.avg_moving_pace()
    
    def compute_points_pace(self) -> None:
        """
        Compute pace at each track point.
        """
        self.gpx.compute_points_pace()

    def min_pace(self) -> float:
        """
        Return the minimum pace during the activity.

        Returns
        -------
        float
            Minimum pace.
        """
        return self.gpx.min_pace()

    def max_pace(self) -> float:
        """
        Return the maximum pace during the activity.

        Returns
        -------
        float
            Maximum pace.
        """
        return self.gpx.max_pace()
    
    def compute_points_ascent_speed(self) -> None:
        """
        Compute ascent speed (kilometers per hour) at each track point.
        """
        self.gpx.compute_points_ascent_speed()

    def min_ascent_speed(self) -> float:
        """
        Return the minimum ascent speed (kilometers per hour) during the activity.

        Returns
        -------
        float
            Minimum ascent speed.
        """
        return self.gpx.min_ascent_speed()

    def max_ascent_speed(self) -> float:
        """
        Return the maximum ascent speed (kilometers per hour) during the activity.

        Returns
        -------
        float
            Maximum ascent speed.
        """
        return self.gpx.max_ascent_speed()

    def remove_metadata(self):
        """
        Remove metadata (ie: metadata will not be written when saving the GPX object as a .gpx file).
        """
        self.gpx_writer.metadata = False

    def remove_elevation(self):
        """
        Remove elevation data (ie: elevation data will not be written when saving the GPX object as a .gpx file).
        """
        self.gpx_writer.ele = False

    def remove_time(self):
        """
        Remove time data (ie: time data will not be written when saving the GPX object as a .gpx file).
        """
        self.gpx_writer.time = False

    def remove_gps_errors(self):
        """
        Remove GPS errors.
        """
        self.gpx.remove_gps_errors()

    def remove_close_points(self, min_dist: float = 1, max_dist: float = 10):
        """
        Remove points that are to close together.

        Parameters
        ----------
        min_dist : float, optional
            Minimal distance between two points, by default 1
        max_dist : float, optional
            Maximal distance between two points, by default 10
        """
        self.gpx.remove_close_points(min_dist, max_dist)

    def simplify(self, tolerance: float = 2):
        """
        Simplify the tracks using Ramer-Douglas-Peucker algorithm.

        Parameters
        ----------
        tolerance : float, optional
            Tolerance (meters). Corresponds to the minimum distance between the
            point and the track before the point is removed, by default 2
        """
        epsilon = degrees(tolerance/EARTH_RADIUS)
        self.gpx.simplify(epsilon)

    def merge(self, gpx: GPX):
        """
        _summary_

        Parameters
        ----------
        gpx : GPX
            _description_
        """
        # TODO
        if self.gpx.tag is None:
            self.gpx.tag = gpx.gpx.tag
        if self.gpx.creator is None:
            self.gpx.creator = gpx.gpx.creator
        if self.gpx.xmlns is None:
            self.gpx.xmlns = gpx.gpx.xmlns
        if self.gpx.version is None:
            self.gpx.version = gpx.gpx.version
        if self.gpx.xmlns_xsi is None:
            self.gpx.xmlns_xsi = gpx.gpx.xmlns_xsi
        self.gpx.xsi_schema_location.extend(gpx.gpx.xsi_schema_location)
        if self.gpx.xmlns_gpxtpx is None:
            self.gpx.xmlns_gpxtpx = gpx.gpx.xmlns_gpxtpx
        if self.gpx.xmlns_gpxx is None:
            self.gpx.xmlns_gpxx = gpx.gpx.xmlns_gpxx
        if self.gpx.xmlns_gpxtrk is None:
            self.gpx.xmlns_gpxtrk = gpx.gpx.xmlns_gpxtrk
        if self.gpx.xmlns_wptx1 is None:
            self.gpx.xmlns_wptx1 = gpx.gpx.xmlns_wptx1
        if self.gpx.metadata is None:
            self.gpx.metadata = gpx.gpx.metadata
        self.gpx.wpt.extend(gpx.gpx.wpt)
        self.gpx.rte.extend(gpx.gpx.rte)
        self.gpx.tracks.extend(gpx.gpx.tracks)
        if self.gpx.extensions is None:
            self.gpx.extensions = gpx.gpx.extensions

    def to_gpx_string(self) -> str:
        """
        Convert the GPX object to a string.

        Returns
        -------
        str
            String representingth GPX object.
        """
        return self.gpx_writer.gpx_to_string(self.gpx)
    
    def to_dataframe(
            self,
            projection: bool = False,
            elevation: bool = True,
            speed: bool = False,
            pace: bool = False,
            ascent_rate: bool = False,
            ascent_speed: bool = False) -> pd.DataFrame:
        """
        Convert GPX object to Pandas Dataframe.

        Parameters
        ----------
        projection : bool, optional
            Toggle projection, by default False
        elevation : bool, optional
            Toggle elevation, by default True
        speed : bool, optional
            Toggle speed, by default False
        pace : bool, optional
            Toggle pace, by default False
        ascent_rate : bool, optional
            Toggle ascent rate, by default False
        ascent_speed : bool, optional
            Toggle ascent speed, by default False

        Returns
        -------
        pd.DataFrame
            Dataframe containing data from GPX.
        """
        return self.gpx.to_dataframe(projection, elevation, speed, pace, ascent_rate, ascent_speed)

    def to_csv(
            self,
            path: str = None,
            sep: str = ",",
            columns: List[str] = None,
            header: bool = True,
            index: bool = False) -> Union[str, None]:
        """
        Write the GPX object track coordinates to a .csv file.

        Parameters
        ----------
        path : str, optional
            Path to the .csv file, by default None
        sep : str, optional
            Separator, by default ","
        columns : List[str], optional
            List of columns to write, by default None
        header : bool, optional
            Toggle header, by default True
        index : bool, optional
             Toggle index, by default False

        Returns
        -------
        str
            CSV like string if path is set to None.
        """
        return self.gpx.to_csv(path, sep, columns, header, index)
    
    def to_gpx(
            self,
            path: str,
            xml_schema: bool = True,
            xml_extensions_schemas: bool = False) -> bool:
        """
        Write the GPX object to a .gpx file.

        Parameters
        ----------
        path : str
            Path to the .gpx file.
        xml_schema : bool, optional
            Toggle schema verification after writting, by default True
        xml_extensions_schemas : bool, optional
            Toggle extensions schema verificaton after writing. Requires internet connection and is not guaranted to work, by default False

        Returns
        -------
        bool
            Return False if written file does not follow checked schemas. Return True otherwise.
        """
        return self.gpx_writer.write(path, xml_schema, xml_extensions_schemas)
    
    def to_kml(
            self,
            path: str,
            styles: Optional[List[Tuple[str, Dict]]] = None,
            xml_schema: bool = True) -> bool:
        """
        Write the GPX object to a .kml file.

        Parameters
        ----------
        path : str
            Path to the .gpx file.
        styles : List[Tuple[str, Dict]], optional
            List of (style_id, style) tuples, by default None
        xml_schema : bool, optional
            Toggle schema verification after writting, by default True

        Returns
        -------
        bool
            Return False if written file does not follow checked schemas. Return True otherwise.
        """
        return self.kml_writer.write(path, styles, xml_schema)

    def _matplotlib_plot_text(
            self,
            fig: Figure,
            duration: Optional[Tuple[float, float]] = (0, 0),
            distance: Optional[Tuple[float, float]] = (0.5, 0),
            ascent: Optional[Tuple[float, float]] = (1, 0),
            pace: Optional[Tuple[float, float]] = (1, 0),
            speed: Optional[Tuple[float, float]] = (1, 0),
            text_parameters: Optional[dict] = None):
        """
        Plot text on Matplotlib plot.

        Parameters
        ----------
        fig : Figure
            Matplotlib figure.
        duration : Optional[Tuple[float, float]], optional
            Duration text plot coordinates, by default (0, 0)
        distance : Optional[Tuple[float, float]], optional
            Distance text plot coordinates, by default (0.5, 0)
        ascent : Optional[Tuple[float, float]], optional
            Ascent text plot coordinates, by default (1, 0)
        pace : Optional[Tuple[float, float]], optional
            Pace text plot coordinates, by default (1, 0)
        speed : Optional[Tuple[float, float]], optional
            Speed text plot coordinates, by default (1, 0)
        text_parameters : Optional[dict], optional
            Text parameters, by default None
        """
        # Handle text parameters
        if text_parameters is None:
            text_parameters = dict(ha='center', va='center', fontsize=10,
                                   transform=fig.axes[0].transAxes, bbox=dict(facecolor='gray', alpha=0.5))
        else:
            text_parameters["transform"] = fig.axes[0].transAxes

        if duration is not None:
            plt.text(duration[0], duration[1],
                     f"Duration:\n{self.total_elapsed_time()}", **text_parameters)

        if distance is not None:
            plt.text(distance[0], distance[1],
                     f"Distance:\n{self.distance()/1000:.2f} km", **text_parameters)

        if ascent is not None:
            plt.text(ascent[0], ascent[1],
                     f"Ascent:\n{self.ascent():.2f} m", **text_parameters)

        elif pace is not None:
            plt.text(pace[0], pace[1],
                     f"Pace:\n{self.avg_pace():.2f} min/km", **text_parameters)

        elif speed is not None:
            plt.text(speed[0], speed[1],
                     f"Speed:\n{self.avg_speed():.2f} km/h", **text_parameters)

    def matplotlib_axes_plot(
            self,
            axes: Axes,
            projection: Optional[str] = None,
            color: str = "#101010",
            colorbar: bool = False,
            start_stop_colors: Optional[Tuple[str, str]] = None,
            way_points_color: Optional[str] = None,
            title: Optional[str] = None,
            duration: Optional[Tuple[float, float]] = None,
            distance: Optional[Tuple[float, float]] = None,
            ascent: Optional[Tuple[float, float]] = None,
            pace: Optional[Tuple[float, float]] = None,
            speed: Optional[Tuple[float, float]] = None):
        """
        Plot GPX on Matplotlib axes.

        Parameters
        ----------
        axes : Axes
            Axes to plot on.
        projection : Optional[str], optional
            Projection, by default None
        color : str, optional
            A color string (ie: "#FF0000" or "red") or a track attribute
            ("elevation", "speed", "pace", "vertical_drop", "ascent_rate",
            "ascent_speed"), by default "#101010"
        colorbar : bool, optional
            Colorbar Toggle, by default False
        start_stop_colors : Optional[Tuple[str, str]], optional
            Start and stop points colors, by default None
        way_points_color : Optional[str], optional
            Way point color, by default None
        title : Optional[str], optional
            Title, by default None
        duration : Optional[Tuple[float, float]], optional
            Display duration, by default None
        distance : Optional[Tuple[float, float]], optional
            Display distance, by default None
        ascent : Optional[Tuple[float, float]], optional
            Display ascent, by default None
        pace : Optional[Tuple[float, float]], optional
            Display pace, by default None
        speed : Optional[Tuple[float, float]], optional
            Display speed, by default None
        """
        # Clear axes
        axes.clear()

        # Handle projection (select dataframe columns to use and project if needed)
        if projection is None:
            column_x = "lon"
            column_y = "lat"
        else:
            column_x = "x"
            column_y = "y"
            self.gpx.project(projection)  # Project all track points

        # Create dataframe containing data from the GPX file
        gpx_df = self.to_dataframe(projection=True, elevation=True, speed=True, pace=True, ascent_rate=True, ascent_speed=True)

        # Scatter all track points
        if color == "elevation":
            cmap = matplotlib.colors.LinearSegmentedColormap.from_list("", ["green","blue"])
            im = axes.scatter(gpx_df[column_x], gpx_df[column_y],
                        c=gpx_df["ele"], cmap=cmap)
        elif color == "speed":
            # cmap = matplotlib.colors.LinearSegmentedColormap.from_list("", ["lightskyblue", "deepskyblue", "blue", "mediumblue", "midnightblue"])
            cmap = matplotlib.colors.LinearSegmentedColormap.from_list("", ["lightskyblue", "mediumblue", "midnightblue"])
            im = axes.scatter(gpx_df[column_x], gpx_df[column_y],
                        c=gpx_df["speed"], cmap=cmap)
        elif color == "pace":
            cmap = matplotlib.colors.LinearSegmentedColormap.from_list("", ["lightskyblue", "midnightblue"])
            im = axes.scatter(gpx_df[column_x], gpx_df[column_y],
                        c=gpx_df["pace"], cmap=cmap)
        elif color == "vertical_drop":
            cmap = matplotlib.colors.LinearSegmentedColormap.from_list("", ["yellow", "orange", "red", "purple", "black"])
            im = axes.scatter(gpx_df[column_x], gpx_df[column_y],
                        c=abs(gpx_df["ascent_rate"]), cmap=cmap)
        elif color == "ascent_rate":
            cmap = matplotlib.colors.LinearSegmentedColormap.from_list("", ["darkgreen", "green", "yellow", "red", "black"])
            im = axes.scatter(gpx_df[column_x], gpx_df[column_y],
                        c=gpx_df["ascent_rate"], cmap=cmap)
        elif color == "ascent_speed":
            cmap = matplotlib.colors.LinearSegmentedColormap.from_list("", ["deeppink", "lightpink", "lightcoral", "red", "darkred"])
            im = axes.scatter(gpx_df[column_x], gpx_df[column_y],
                        c=gpx_df["ascent_speed"], cmap=cmap)
        else:
            im = axes.scatter(gpx_df[column_x], gpx_df[column_y], color=color)
        
        # Colorbar
        if colorbar:
            plt.colorbar(im)

        # Scatter start and stop points with different color
        if start_stop_colors is not None:
            axes.scatter(gpx_df[column_x][0],
                        gpx_df[column_y][0], color=start_stop_colors[0])
            axes.scatter(gpx_df[column_x][len(gpx_df[column_x])-1],
                        gpx_df[column_y][len(gpx_df[column_x])-1], color=start_stop_colors[1])

        # Scatter way points with different color
        if way_points_color:
            for way_point in self.gpx.wpt:
                if projection:
                    way_point.project(projection)
                    axes.scatter(way_point._x, way_point._y, color=way_points_color)
                else:
                    axes.scatter(way_point.lon, way_point.lat, color=way_points_color)

        # Add title
        if title is not None:
            axes.set_title(title, size=20)

        # Add text elements
        self._matplotlib_plot_text(axes.get_figure(), duration, distance, ascent, pace, speed)

        # Add ticks
        axes.set_xticks([min(gpx_df[column_x]), max(gpx_df[column_x])])
        axes.set_yticks([min(gpx_df[column_y]), max(gpx_df[column_y])])

        # Add axis limits (useless??)
        if projection is not None:
            axes.set_xlim(left=min(gpx_df[column_x]),
                        right=max(gpx_df[column_x]))
            axes.set_ylim(bottom=min(gpx_df[column_y]),
                        top=max(gpx_df[column_y]))

    def matplotlib_plot(
        self,
        projection: Optional[str] = None,
        color: str = "#101010",
        colorbar: bool = False,
        start_stop_colors: Optional[Tuple[str, str]] = None,
        way_points_color: Optional[str] = None,
        title: Optional[str] = None,
        duration: Optional[Tuple[float, float]] = None,
        distance: Optional[Tuple[float, float]] = None,
        ascent: Optional[Tuple[float, float]] = None,
        pace: Optional[Tuple[float, float]] = None,
        speed: Optional[Tuple[float, float]] = None,
        file_path: Optional[str] = None):
        """
        Plot GPX using Matplotlib.

        Parameters
        ----------
        projection : Optional[str], optional
            Projection, by default None
        color : str, optional
            A color string (ie: "#FF0000" or "red") or a track attribute
            ("elevation", "speed", "pace", "vertical_drop", "ascent_rate",
            "ascent_speed"), by default "#101010"
        colorbar : bool, optional
            Colorbar Toggle, by default False
        start_stop_colors : Optional[Tuple[str, str]], optional
            Start and stop points colors, by default None
        way_points_color : Optional[str], optional
            Way point color, by default None
        title : Optional[str], optional
            Title, by default None
        duration : Optional[Tuple[float, float]], optional
            Display duration, by default None
        distance : Optional[Tuple[float, float]], optional
            Display distance, by default None
        ascent : Optional[Tuple[float, float]], optional
            Display ascent, by default None
        pace : Optional[Tuple[float, float]], optional
            Display pace, by default None
        speed : Optional[Tuple[float, float]], optional
            Display speed, by default None
        """
        # Create figure with axes
        fig = plt.figure(figsize=(14, 8))
        fig.add_subplot(111)

        # Plot on axes
        self.matplotlib_axes_plot(fig.axes[0],
                                  projection,
                                  color,
                                  colorbar,
                                  start_stop_colors,
                                  way_points_color,
                                  title,
                                  duration,
                                  distance,
                                  ascent,
                                  pace,
                                  speed)
        
        # Save or display plot
        if file_path is not None:
            directory_path = os.path.dirname(os.path.realpath(file_path))
            if not os.path.exists(directory_path):
                logging.error("Provided path does not exist")
                return
            plt.savefig(file_path)
        else:
            plt.show()

    def matplotlib_basemap_plot(
            self,
            projection: str = "cyl",
            service: str = "World_Shaded_Relief",
            color: str = "#101010",
            start_stop_colors: Optional[Tuple[str, str]] = None,
            way_points_color: Optional[str] = None,
            title: Optional[str] = None,
            duration: Optional[Tuple[float, float]] = None,
            distance: Optional[Tuple[float, float]] = None,
            ascent: Optional[Tuple[float, float]] = None,
            pace: Optional[Tuple[float, float]] = None,
            speed: Optional[Tuple[float, float]] = None,
            file_path: str = None):
        """
        Plot GPX using Matplotlib Basemap Toolkit.

        Parameters
        ----------
        projection : str, optional
            Projection, by default "cyl"
        service : str, optional
            Service used to fetch map background. Currently supported services:
            "World_Shaded_Relief". Defaults to "World_Shaded_Relief",
            by default "World_Shaded_Relief"
        color : str, optional
            Track color, by default "#101010"
        start_stop_colors : Optional[Tuple[str, str]], optional
            Start and stop points colors, by default None
        way_points_color : Optional[str], optional
            Way point color, by default None
        title : Optional[str], optional
            Title, by default None
        duration : Optional[Tuple[float, float]], optional
            Display duration, by default None
        distance : Optional[Tuple[float, float]], optional
            Display distance, by default None
        ascent : Optional[Tuple[float, float]], optional
            Display ascent, by default None
        pace : Optional[Tuple[float, float]], optional
            Display pace, by default None
        speed : Optional[Tuple[float, float]], optional
            Display speed, by default None
        file_path : str, optional
            Path to save plot, by default None
        """
        # Create figure
        fig = plt.figure(figsize=(14, 8))

        # Create square map
        center_lat, center_lon = self.center()
        min_lat, min_lon, max_lat, max_lon = self.bounds()
        offset = 0.001
        min_lat, min_lon = max(0, min_lat - offset), max(0, min_lon - offset)
        max_lat, max_lon = min(max_lat + offset, 90), min(max_lon + offset, 90)

        delta_lat = max_lat - min_lat
        delta_lon = max_lon - min_lon

        if delta_lat > delta_lon:
            min_lon = max(0, center_lon - delta_lat/2)
            max_lon = min(90, center_lon + delta_lat/2)
        else:
            min_lat = max(0, center_lat - delta_lon/2)
            max_lat = min(90, center_lat + delta_lon/2)

        map = Basemap(projection=projection,
                      llcrnrlon=min_lon,
                      llcrnrlat=min_lat,
                      urcrnrlon=max_lon,
                      urcrnrlat=max_lat)

        # map.drawmapboundary(fill_color='aqua')
        # map.fillcontinents(color='coral',lake_color='aqua')
        # map.drawcoastlines()

        # map.bluemarble()
        # map.shadedrelief()
        # map.etopo()
        map.arcgisimage(service=service,
                        xpixels=2000, dpi=200, verbose=True)

        # Create dataframe containing data from the GPX file
        gpx_df = self.to_dataframe()

        # Project track points
        x, y = map(gpx_df["lon"], gpx_df["lat"])
        
        # Plot track points on the map
        x = x.tolist()
        y = y.tolist()
        map.plot(x, y, marker=None, color=color)

        # Scatter start and stop points with different color
        if start_stop_colors:
            map.plot(x[0], y[0], marker="^", color=start_stop_colors[0])
            map.plot(x[-1], y[-1], marker="h", color=start_stop_colors[1])

        # Project and scatter way points with different color
        if way_points_color:
            for way_point in self.gpx.wpt:
                x, y = map(way_point.lon, way_point.lat)
                map.plot(x, y, marker="D", color="waypoint_color")

        # Add title
        if title is not None:
            plt.title(title, size=20)

        # Add text elements
        self._matplotlib_plot_text(fig, duration, distance, ascent, pace, speed)

        # Save or display plot
        if file_path is not None:
            # Check if provided path exists
            directory_path = os.path.dirname(os.path.realpath(file_path))
            if not os.path.exists(directory_path):
                logging.error("Provided path does not exist")
                return
            plt.savefig(file_path)
        else:
            plt.show()

    def gmplot_plot(
            self,
            color: str = "#110000",
            start_stop_colors: Optional[Tuple[str, str]] = None,
            way_points_color: Optional[str] = None,
            zoom: float = 10.0,
            title: Optional[str] = None,
            file_path: str = None,
            open: bool = True,
            scatter: bool = False,
            plot: bool = True):
        """
        Plot GPX using gmplot.

        Parameters
        ----------
        color : str, optional
            Track color, by default "#110000"
        start_stop_colors : Optional[Tuple[str, str]], optional
            Start and stop points colors, by default None
        way_points_color : Optional[str], optional
            Way points color, by default None
        zoom : float, optional
            Zoom, by default 10.0
        title : Optional[str], optional
            Title, by default None
        file_path : str, optional
            Path to save plot, by default None
        open : bool, optional
            Open the plot in the default web browser, by default True
        scatter : bool, optional
            Scatter track points, by default False
        plot : bool, optional
            Plot track points, by default True
        """
        # Create plotter
        center_lat, center_lon = self.center()
        map = gmplot.GoogleMapPlotter(center_lat, center_lon, zoom)

        # Create dataframe containing data from the GPX file
        gpx_df = self.to_dataframe()

        # Scatter track points
        if scatter:
            map.scatter(gpx_df["lat"], gpx_df["lon"],
                        color, size=5, marker=False)
        if plot:
            map.plot(gpx_df["lat"], gpx_df["lon"],
                     color, edge_width=2.5)
        
        # Scatter start and stop points with different color
        if start_stop_colors:
            map.scatter([self.gpx.tracks[0].trkseg[0].trkpt[0].lat],
                        [self.gpx.tracks[0].trkseg[0].trkpt[0].lon],
                        start_stop_colors[0], size=5, marker=True)
            map.scatter([self.gpx.tracks[-1].trkseg[-1].trkpt[-1].lat],
                        [self.gpx.tracks[-1].trkseg[-1].trkpt[-1].lon],
                        start_stop_colors[1], size=5, marker=True)
            
        # Scatter way points with different color
        if way_points_color:
            for way_point in self.gpx.wpt:
                map.scatter([way_point.lat], [way_point.lon],
                            way_points_color, size=5, marker=True)
        
        # Add title
        if title is not None:
            map.text(center_lat, center_lon, self.name(), color="#FFFFFF")

        # Save map
        map.draw(file_path)

        # Open map in web browser
        if open:
            webbrowser.open(file_path)

    def folium_plot(
            self,
            tiles: str = "OpenStreetMap",  # "OpenStreetMap", "Stamen Terrain", "Stamen Toner"
            color: str = "#110000",
            start_stop_colors: Optional[Tuple[str, str]] = None,
            way_points_color: Optional[str] = None,
            minimap: bool = False,
            coord_popup: bool = False,
            title: Optional[str] = None,
            zoom: float = 12.0,
            file_path: str = None,
            open: bool = True):
        """
        Plot GPX using folium.

        Parameters
        ----------
        tiles : str, optional
            Map tiles. Supported tiles: "OpenStreetMap", "Stamen Terrain",
            "Stamen Toner", by default "OpenStreetMap"
        start_stop_colors : Optional[Tuple[str, str]], optional
            Start and stop points colors, by default None
        way_points_color : Optional[str], optional
            Way points color, by default None
        minimap : bool, optional
            Add minimap, by default False
        coord_popup : bool, optional
            Add coordinates pop-up when clicking on the map, by default False
        title : Optional[str], optional
            Title, by default None
        zoom : float, optional
            Zoom, by default 12.0
        file_path : str, optional
            Path to save plot, by default None
        open : bool, optional
            Open the plot in the default web browser, by default True
        """
        # Create map
        center_lat, center_lon = self.center()
        m = folium.Map(location=[center_lat, center_lon],
                       zoom_start=zoom,
                       tiles=tiles)
        
        # Plot track points
        gpx_df = self.to_dataframe()
        gpx_df["coordinates"] = list(
            zip(gpx_df.lat, gpx_df.lon))
        folium.PolyLine(gpx_df["coordinates"],
                        tooltip=self.name(), color=color).add_to(m)

        # Scatter start and stop points with different color
        if start_stop_colors:
            folium.Marker([self.gpx.tracks[0].trkseg[0].trkpt[0].lat, self.gpx.tracks[0].trkseg[0].trkpt[0].lon],
                          popup="<b>Start</b>", tooltip="Start", icon=folium.Icon(color=start_stop_colors[0])).add_to(m)
            folium.Marker([self.gpx.tracks[-1].trkseg[-1].trkpt[-1].lat, self.gpx.tracks[-1].trkseg[-1].trkpt[-1].lon],
                          popup="<b>Stop</b>", tooltip="Stop", icon=folium.Icon(color=start_stop_colors[1])).add_to(m)
        
        # Scatter way points with different color
        if way_points_color:
            for way_point in self.gpx.wpt:
                folium.Marker([way_point.lat, way_point.lon], popup="<i>Way point</i>",
                              tooltip="Way point", icon=folium.Icon(icon="info-sign", color=way_points_color)).add_to(m)

        # Add minimap
        if minimap:
            minimap = MiniMap(toggle_display=True)
            minimap.add_to(m)

        # Add latitude-longitude pop-up
        if coord_popup:
            m.add_child(folium.LatLngPopup())

        # Title
        if title is not None:
            folium.map.Marker(
                [center_lat, center_lon],
                icon=DivIcon(
                    icon_size=(250, 36),
                    icon_anchor=(0, 0),
                    html=f'<div style="font-size: 20pt">{title}</div>',
                )
            ).add_to(m)

        # Save map
        if file_path is None:
            file_path = "unnamed.html"
        m.save(file_path)

        # Open map in web browser
        if open:
            webbrowser.open(file_path)

    def write_tmp_kml(
            self,
            path: str ="tmp.kml",
            kml_string: Optional[bytes] = None):
        """
        Write temproray .KML file in order to parse KMZ file.
        """
        # Open/create KML file
        try:
            f = open(path, "wb")
        except OSError:
            logging.exception(f"Could not open/read file: {path}")
            raise
        # Write KML file
        with f:
            if kml_string is not None:
                f.write(kml_string)



class Bounds():
    """
    boundsType element in GPX file.
    """

    def __init__(
            self,
            tag: str = "bounds",
            minlat: float = None,
            minlon: float = None,
            maxlat: float = None,
            maxlon: float = None) -> None:
        """
        Initialize Bounds instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "bounds"
        minlat : float, optional
            Minimal latitude, by default None
        minlon : float, optional
            Minimal longitude, by default None
        maxlat : float, optional
            Maximal latitude, by default None
        maxlon : float, optional
            Maximal longitude, by default None
        """
        self.tag: str = tag
        self.minlat: float = minlat
        self.minlon: float = minlon
        self.maxlat: float = maxlat
        self.maxlon: float = maxlon


import sys
import os
import pytest
import filecmp
from shutil import rmtree

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

file_folder = os.path.dirname(__file__)
parent_folder = os.path.realpath(os.path.dirname(file_folder)) # ie: ezGPX

os.chdir(file_folder)
sys.path.append(parent_folder + "/ezgpx")

from ezgpx import GPX

FILES_DIRECTORY = "test_files/files/"
REFERENCE_FILES_DIRECTORY = "test_files/reference_files/"

class TestGPX():

    #==== Init ===============================================================#

    def test_init(self):
        # Create temporary folder
        rmtree("tmp", True)
        os.makedirs(os.path.dirname(__file__) + "/tmp")

    #==== Parsing/Writing ====================================================#

    def test_parse(self):
        # Parse GPX files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"), xml_schema=True, xml_extensions_schemas=False)
        invalid_gpx = GPX(os.path.join(FILES_DIRECTORY, "invalid_schema.gpx"), xml_schema=False, xml_extensions_schemas=False)

    #==== Check Schemas ======================================================#check_xml_schemas

    def test_check_schemas(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"), xml_schema=False, xml_extensions_schemas=False)
        invalid_gpx = GPX(os.path.join(FILES_DIRECTORY, "invalid_schema.gpx"), xml_schema=False, xml_extensions_schemas=False)
        # Tests
        assert(gpx.check_xml_schema() == True)
        assert(invalid_gpx.check_xml_schema() == False)

    #==== Properties =========================================================#

    def test_file_name(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.file_name() == "strava_run_1.gpx")

    def test_name(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.name() == "Dérouillage habituel 💥")

    def test_set_name(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        new_name = "test"
        gpx.set_name(new_name)
        assert(gpx.name() == new_name)

    def test_nb_points(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.nb_points() == 939)

    @pytest.mark.skip(reason="nothing to test")
    def test_first_point(self):
        pass

    @pytest.mark.skip(reason="nothing to test")
    def test_last_point(self):
        pass

    def test_bounds(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.bounds() == (44.032965, 4.444134, 44.047778, 4.486607))

    def test_center(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.center() == (44.0403715, 4.465370500000001))

    def test_distance(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.distance() == 10922.788757238777)

    def test_ascent(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.ascent() == 225.29999999999995)

    def test_descent(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.descent() == 224.79999999999987)

    @pytest.mark.skip(reason="nothing to test")
    def test_compute_points_ascent_rate(self):
        pass

    def test_min_ascent_rate(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.min_ascent_rate() == -38.54138626353898)

    def test_max_ascent_rate(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.max_ascent_rate() == 51.2919872933083)

    def test_min_elevation(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.min_elevation() == 98.5)

    def test_max_elevation(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.max_elevation() == 235.6)

    @pytest.mark.skip(reason="time related test")
    def test_start_time(self):
        pass

    @pytest.mark.skip(reason="time related test")
    def test_stop_time(self):
        pass

    @pytest.mark.skip(reason="time related test")
    def test_total_elapsed_time(self):
        pass

    @pytest.mark.skip(reason="time related test")
    def test_stopped_time(self):
        pass

    @pytest.mark.skip(reason="time related test")
    def test_moving_time(self):
        pass

    def test_avg_speed(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.avg_speed() == 10.66505004775145)

    def test_avg_moving_speed(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.avg_moving_speed() == 10.974613320139435)

    @pytest.mark.skip(reason="nothing to test")
    def test_compute_points_speed(self):
        pass

    def test_min_speed(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.min_speed() == 0.0)

    def test_max_speed(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.max_speed() == 19.69510525631602)

    def test_avg_pace(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.avg_pace() == 5.625852643105975)

    def test_avg_moving_pace(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.avg_moving_pace() == 5.467163010645161)

    @pytest.mark.skip(reason="nothing to test")
    def test_compute_points_pace(self):
        pass
    
    def test_min_pace(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.min_pace() == 3.046442210851278)

    def test_max_pace(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.max_pace() == 1041.2956304834424)

    @pytest.mark.skip(reason="nothing to test")
    def test_compute_points_ascent_speed(self):
        pass

    def test_min_ascent_speed(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.min_ascent_speed() == -3.3599999999999794)

    def test_max_ascent_speed(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert(gpx.max_ascent_speed() == 2.1599999999999797)

    #==== Modifications ======================================================#

    #==== Conversion and Saving ==============================================#

    def test_to_dataframe(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        df = gpx.to_dataframe()
        reference_df = pd.read_csv(os.path.join(FILES_DIRECTORY, "strava_run_1_test.csv"))
        # Test
        assert(reference_df.equals(df))

    def test_to_gpx(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        gpx.to_gpx("tmp/strava_run_1.gpx", xml_schema=False)
        # Test
        assert(filecmp.cmp("tmp/strava_run_1.gpx", os.path.join(REFERENCE_FILES_DIRECTORY, "strava_run_1_test.gpx"), False))
        
    def test_to_kml(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        gpx.to_kml("tmp/strava_run_1_test.kml", styles=None, xml_schema=False)
        # Test
        assert(filecmp.cmp("tmp/strava_run_1_test.kml", os.path.join(REFERENCE_FILES_DIRECTORY, "strava_run_1_test.kml"), False))

    def test_to_csv(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        gpx.to_csv("tmp/strava_run_1.csv", columns=["lat", "lon", "ele"])
        # Test
        assert(filecmp.cmp("tmp/strava_run_1.csv", os.path.join(REFERENCE_FILES_DIRECTORY, "strava_run_1_test.csv"), False))

    #==== Plots ==============================================================#

    def _test_matplotlib_plot_1(self):
        # Plot
        self.gpx.matplotlib_plot(start_stop_colors=None, color="#ffffff", title="Track", file_path="tmp/matplotlib_strava_run_1.png")
        # Load images
        test_img = plt.imread("tmp/matplotlib_strava_run_1.png")
        ref_img = plt.imread(os.path.join(REFERENCE_FILES_DIRECTORY, "matplotlib_strava_run_1.png"))
        # Compare images
        return np.array_equal(test_img, ref_img)
    
    def _test_matplotlib_plot_2(self):
        # Plot
        self.gpx.matplotlib_plot(start_stop_colors=("#00FF00", "#FF0000"), color="#ffffff", title="Track", file_path="tmp/matplotlib_strava_run_1_start_stop.png")
        # Load images
        test_img = plt.imread("tmp/matplotlib_strava_run_1_start_stop.png")
        ref_img = plt.imread(os.path.join(REFERENCE_FILES_DIRECTORY, "matplotlib_strava_run_1_start_stop.png"))

        # Compare images
        return np.array_equal(test_img, ref_img)
    
    def _test_matplotlib_plot_3(self):
        # Plot
        self.gpx.matplotlib_plot(start_stop_colors=None, color=True, title="Track", file_path="tmp/matplotlib_strava_run_1_elevation.png")
        # Load images
        test_img = plt.imread("tmp/matplotlib_strava_run_1_elevation.png")
        ref_img = plt.imread(os.path.join(REFERENCE_FILES_DIRECTORY, "matplotlib_strava_run_1_elevation.png"))
        # Compare images
        return np.array_equal(test_img, ref_img)
    
    def _test_matplotlib_plot_4(self):
        # Plot
        self.gpx.matplotlib_plot(start_stop_colors=("#00FF00", "#FF0000"), color=True, title="Track", file_path="tmp/matplotlib_strava_run_1_start_stop_elevation.png")
        # Load images
        test_img = plt.imread("tmp/matplotlib_strava_run_1_start_stop_elevation.png")
        ref_img = plt.imread(os.path.join(REFERENCE_FILES_DIRECTORY, "matplotlib_strava_run_1_start_stop_elevation.png"))
        # Compare images
        return np.array_equal(test_img, ref_img)

    @pytest.mark.skip(reason="not ready")
    def test_matplotlib_plot(self):
        # Parse GPX file
        self.gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Tests
        assert(self._test_matplotlib_plot_1())
        assert(self._test_matplotlib_plot_2())
        assert(self._test_matplotlib_plot_3())
        assert(self._test_matplotlib_plot_4())
        # self._test_matplotlib_plot_1()
        # self._test_matplotlib_plot_2()
        # self._test_matplotlib_plot_3()
        # self._test_matplotlib_plot_4()

    def _test_folium_plot_1(self):
        # Plot
        self.gpx.folium_plot(tiles="openStreetMap",
                             color="#110000",
                             start_stop_colors=None,
                             way_points_color=None,
                             minimap=False,
                             coord_popup=False,
                             title=None,
                             zoom=12,
                             file_path="tmp/folium_strava_run_1.html",
                             open=False)
        # Compare files
        return filecmp.cmp("tmp/folium_strava_run_1.html", os.path.join(REFERENCE_FILES_DIRECTORY, "folium_strava_run_1.html"), False)

    @pytest.mark.skip(reason="not ready")
    def test_folium_plot(self):
        self.test_init() # For developping purpose only (using: pytest test_GPX.py::TestGPX::test_folium_plot)
        # Parse GPX file
        self.gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Tests
        assert(self._test_folium_plot_1())
        # assert(self._test_folium_plot_2())
        # assert(self._test_folium_plot_3())
        # assert(self._test_folium_plot_4())
        # self._test_folium_plot_1()
        # self._test_folium_plot_2()à
        # self._test_folium_plot_3()
        # self._test_folium_plot_4()

    #==== Destroy ============================================================#

    def test_destroy(self, remove_tmp: bool = True):
        # Remove temporary folder
        if remove_tmp:
            rmtree("tmp", True)

    #==== Test ===============================================================#

    @pytest.mark.skip(reason="test") # https://docs.pytest.org/en/7.3.x/how-to/skipping.html
    def test_test(self, remove_tmp: bool = True):
         # Create temporary folder
        rmtree("tmp", True)
        os.makedirs(os.path.dirname(__file__) + "/tmp")
        # Parse GPX file
        self.gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Remove temporary folder
        if remove_tmp:
            rmtree("tmp")


import os
from typing import Optional, Union
import logging
from datetime import datetime
import xml.etree.ElementTree as ET

from ..xml_parser import XMLParser
from ..gpx_elements import Bounds, Copyright, Email, Extensions, Gpx, Link, Metadata, Person, Point, PointSegment, Route, TrackSegment, Track, WayPoint

class GPXParser(XMLParser):
    """
    GPX file parser.
    """

    def __init__(
            self,
            file_path: Optional[str] = None,
            check_xml_schemas: bool = True,
            xml_extensions_schemas: bool = False) -> None:
        """
        Initialize GPXParser instance.

        Args:
            file_path (str, optional): Path to the file to parse. Defaults to None.
            check_xml_schemas (bool, optional): Toggle schema verification during parsing. Defaults to True.
            extensions_schemas (bool, optional): Toggle extensions schema verificaton durign parsing. Requires internet connection and is not guaranted to work. Defaults to False.
        """
        if not file_path.endswith(".gpx"):
            return
        super().__init__(file_path,
                         {"topo": "http://www.topografix.com/GPX/1/1"},
                         check_xml_schemas, 
                         xml_extensions_schemas)
 
        if self.file_path is not None and os.path.exists(self.file_path):
            self.parse()
        else:
            logging.warning("File path does not exist")

    def find_precisions(self):
        """
        Find decimal precision of any type of value in a GPX file (latitude, elevation...).
        """
        # Point
        track = self.xml_root.findall("topo:trk", self.name_space)[0]
        segment = track.findall("topo:trkseg", self.name_space)[0]
        point = segment.findall("topo:trkpt", self.name_space)[0]

        self.precisions["lat_lon"] = self.find_precision(point.get("lat"))
        self.precisions["elevation"] = self.find_precision(self.find_text(point, "topo:ele"))

        # Extensions
        # extensions = segment.find("topo:extensions", self.name_space)

        # if extensions is not None:
        #     self.precisions["distance"] = self.find_precision(self.find_text(extensions, "topo:Distance"))
        #     self.precisions["duration"] = self.find_precision(self.find_text(extensions, "topo:TotalElapsedTime"))
        #     self.precisions["speed"] = self.find_precision(self.find_text(extensions, "topo:MovingSpeed"))
        #     self.precisions["rate"] = self.find_precision(self.find_text(extensions, "topo:AvgAscentRate"))

    def find_time_element(self) -> Union[str, None]:
        """
        Find a time element in GPX file.

        Returns:
            Union[str, None]: Time element.
        """
        # Use time from metadata
        metadata = self.xml_root.find("topo:metadat", self.name_space)
        time = self.find_text(metadata, "topo:time")
        if time is not None:
            return time

        # Use time from track point
        track = self.xml_root.findall("topo:trk", self.name_space)[0]
        segment = track.findall("topo:trkseg", self.name_space)[0]
        point = segment.findall("topo:trkpt", self.name_space)[0]
        time = self.find_text(point, "topo:time")
        if time is not None:
            return time
        
        # No time element at all...
        return None

    def find_time_format(self):
        """
        Find the time format used in GPX file. 
        """
        time = self.find_time_element()
        if time is None:
            logging.warning("No time element in GPX file.")
            return

        try:
            d = datetime.strptime(time, "%Y-%m-%dT%H:%M:%SZ")
            self.time_format = "%Y-%m-%dT%H:%M:%SZ"
        except:
            self.time_format = "%Y-%m-%dT%H:%M:%S.%fZ"

    def parse_bounds(self, bounds, tag: str ="bounds") -> Union[Bounds, None]:
        """
        Parse boundsType element from GPX file.

        Args:
            bounds (xml.etree.ElementTree.Element): Parsed bounds element.
            tag (str, Optional): XML tag. Defaults to "bounds".

        Returns:
            Bounds: Bounds instance.
        """
        if bounds is None:
            return None

        minlat = self.get_text(bounds, "minlat")
        minlon = self.get_text(bounds, "minlon")
        maxlat = self.get_text(bounds, "maxlat")
        maxlon = self.get_text(bounds, "maxlon")

        return Bounds(tag, minlat, minlon, maxlat, maxlon)
    
    def parse_copyright(self, copyright, tag: str ="copyright") -> Union[Copyright, None]:
        """
        Parse copyrightType element from GPX file.

        Args:
            copyright (xml.etree.ElementTree.Element): Parsed copyright element.
            tag (str, Optional): XML tag. Defaults to "copyright".

        Returns:
            Copyright: Copyright instance.
        """
        if copyright is None:
            return None
        
        author = self.get_text(copyright, "author")
        year = self.find_text(copyright, "topo:year")
        licence = self.find_text(copyright, "topo:licence")

        return Copyright(tag, author, year, licence)
    
    def parse_email(self, email, tag: str ="email") -> Union[Email, None]:
        """
        Parse emailType element from GPX file.

        Args:
            email (xml.etree.ElementTree.Element): Parsed email element.
            tag (str, Optional): XML tag. Defaults to "email".

        Returns:
            Email: Email instance.
        """
        if email is None:
            return None
        
        id = self.get_text(email, "id")
        domain = self.get_text(email, "domain")

        return Email(tag, id, domain)
    
    def parse_extensions(self, extensions, tag: str ="extensions") -> Union[Extensions, None]:
        """
        Parse extensionsType element from GPX file.

        Args:
            extensions (xml.etree.ElementTree.Element): Parsed extensions element.
            tag (str, Optional): XML tag. Defaults to "extensions".

        Returns:
            Extensions: Extensions instance.
        """
        if extensions is None:
            return None
        
        # display_color = self.find_text(extensions, "topo:DisplayColor")
        # distance = self.find_text(extensions, "topo:Distance")
        # total_elapsed_time = self.find_text(extensions, "topo:TotalElapsedTime")
        # moving_time = self.find_text(extensions, "topo:MovingTime")
        # stopped_time = self.find_text(extensions, "topo:StoppedTime")
        # moving_speed = self.find_text(extensions, "topo:MovingSpeed")
        # max_speed = self.find_text(extensions, "topo:MaxSpeed")
        # max_elevation = self.find_text(extensions, "topo:MaxElevation")
        # min_elevation = self.find_text(extensions, "topo:MinElevation")
        # ascent = self.find_text(extensions, "topo:Ascent")
        # descent = self.find_text(extensions, "topo:Descent")
        # avg_ascent_rate = self.find_text(extensions, "topo:AvgAscentRate")
        # max_ascent_rate = self.find_text(extensions, "topo:MaxAscentRate")
        # avg_descent_rate = self.find_text(extensions, "topo:AvgDescentRate")
        # max_descent_rate = self.find_text(extensions, "topo:MaxDescentRate")

        # return Extensions(tag, display_color, distance, total_elapsed_time, moving_time, stopped_time, moving_speed, max_speed, max_elevation, min_elevation, ascent, descent, avg_ascent_rate, max_ascent_rate, avg_descent_rate, max_descent_rate)
        return None
    
    def parse_link(self, link, tag: str ="link") -> Union[Link, None]:
        """
        Parse linkType element from GPX file.

        Args:
            link (xml.etree.ElementTree.Element): Parsed link element.
            tag (str, Optional): XML tag. Defaults to "link".

        Returns:
            Link: Link instance.
        """
        if link is None:
            return None
        
        href = self.get_text(link, "href")
        text = self.find_text(link, "topo:text")
        type = self.find_text(link, "topo:type")

        return Link(tag, href, text, type)
    
    def parse_metadata(self, metadata, tag: str = "metadata") -> Union[Metadata, None]:
        """
        Parse metadataType element from GPX file.

        Args:
            metadata (xml.etree.ElementTree.Element): Parsed metadata element.
            tag (str, Optional): XML tag. Defaults to "metadata".

        Returns:
            Metadata: Metadata instance.
        """
        if metadata is None:
            return None

        name = self.find_text(metadata, "topo:name")
        desc = self.find_text(metadata, "topo:desc")
        author = self.parse_person(metadata.find("topo:author", self.name_space))
        copyright = self.parse_copyright(metadata.find("topo:copyright", self.name_space))
        link = self.parse_link(metadata.find("topo:link", self.name_space))
        time = self.find_time(metadata, "topo:time")
        keywords = self.find_text(metadata, "topo:keywords")
        bounds = self.parse_bounds(metadata.find("topo:bounds", self.name_space))
        extensions = self.parse_extensions(metadata.find("topo:extensions", self.name_space))

        return Metadata(tag, name, desc, author, copyright, link, time, keywords, bounds, extensions)

    def parse_person(self, person, tag: str ="person") -> Union[Person, None]:
        """
        Parse personType element from GPX file.

        Args:
            person (xml.etree.ElementTree.Element): Parsed person element.
            tag (str, Optional): XML tag. Defaults to "person".

        Returns:
            Person: Person instance.
        """
        if person is None:
            return None
        
        name = self.find_text(person, "topo:name")
        email = self.parse_email(person.find("topo:email", self.name_space))
        link = self.parse_link(person.find("topo:link", self.name_space))

        return Person(tag, name, email, link)
    
    def parse_point_segment(self, point_segment, tag: str = "ptseg") -> PointSegment:
        pass

    def parse_point(self, point, tag: str = "pt") -> Union[Point, None]:
        """
        Parse ptType element from GPX file.

        Args:
            point (xml.etree.ElementTree.Element): Parsed pt element.
            tag (str, Optional): XML tag. Defaults to "pt".

        Returns:
            Point: Point instance.
        """
        if point is None:
            return None
        
        lat = self.get_float(point, "lat")
        lon = self.get_float(point, "lon")
        ele = self.find_float(point, "topo:ele")
        time = self.find_time(point, "topo:time")

        return Point(tag, lat, lon, ele, time)
    
    def parse_route(self, route, tag: str = "rte") -> Union[Route, None]:
        """
        Parse rteType element from GPX file.

        Args:
            route (xml.etree.ElementTree.Element): Parsed rte element.
            tag (str, Optional): XML tag. Defaults to "rte".

        Returns:
            Route: Route instance.
        """
        if route is None:
            return None
        
        name = self.find_text(route, "topo:name")
        cmt = self.find_text(route, "topo:cmt")
        desc = self.find_text(route, "topo:desc")
        src = self.find_text(route, "topo:src")
        link = self.parse_link(route.find("topo:link", self.name_space))
        number = self.find_int(route, "topo:number")
        type = self.find_text(route, "topo:type")
        extensions = self.parse_extensions(route.find("topo:extensions", self.name_space))

        rtept = []
        way_points = route.findall("topo:rtept", self.name_space)
        for way_point in way_points:
            rtept.append(self.parse_way_point(way_point))

        return Route(tag, name, cmt, desc, src, link, number, type, extensions, rtept)

    def parse_track_segment(self, track_segment, tag: str = "trkseg") -> Union[TrackSegment, None]:
        """
        Parse trksegType element from GPX file.

        Args:
            track_segment (xml.etree.ElementTree.Element): Parsed trkseg element.
            tag (str, Optional): XML tag. Defaults to "trkseg".

        Returns:
            TrackSegment: TrackSegment instance.
        """
        if track_segment is None:
            return None
        
        # Points
        trkpt = []
        track_points = track_segment.findall("topo:trkpt", self.name_space)
        for track_point in track_points:
            trkpt.append(self.parse_way_point(track_point, "trkpt"))

        # Extensions
        extensions = self.parse_extensions(track_segment.find("topo:extensions", self.name_space))

        return TrackSegment(tag, trkpt, extensions)

    def parse_track(self, track, tag: str = "trk") -> Union[Track, None]:
        """
        Parse trkType element from GPX file.

        Args:
            track (xml.etree.ElementTree.Element): Parsed trk element.
            tag (str, Optional): XML tag. Defaults to "trk".

        Returns:
            Track: Track instance.
        """
        if track is None:
            return None
        
        name = self.find_text(track, "topo:name")
        cmt = self.find_text(track, "topo:cmt")
        desc = self.find_text(track, "topo:desc")
        src = self.find_text(track, "topo:src")
        link = self.parse_link(track.find("topo:link", self.name_space))
        number = self.find_int(track, "topo:number")
        type = self.find_text(track, "topo:type")
        extensions = self.parse_extensions(track.find("topo:extensions", self.name_space))

        trkseg = []
        segments = track.findall("topo:trkseg", self.name_space)
        for segment in segments:
            trkseg.append(self.parse_track_segment(segment))

        return Track(tag, name, cmt, desc, src, link, number, type, extensions, trkseg)
    
    def parse_way_point(self, way_point, tag: str = "wpt") -> Union[WayPoint, None]:
        """
        Parse wptType element from GPX file.

        Args:
            way_point (xml.etree.ElementTree.Element): Parsed wpt element.
            tag (str, Optional): XML tag. Defaults to "person".

        Returns:
            WayPoint: WayPoint instance.
        """
        if way_point is None:
            return None
        
        lat = self.get_float(way_point, "lat")
        lon = self.get_float(way_point, "lon")
        ele = self.find_float(way_point, "topo:ele")
        time = self.find_time(way_point, "topo:time")
        mag_var = self.find_float(way_point, "topo:magvar")
        geo_id_height = self.find_float(way_point, "topo:geoidheight")
        geo_id_height = self.find_float(way_point, "topo:geoidheight")
        name = self.find_text(way_point, "topo:name")
        cmt = self.find_text(way_point, "topo:cmt")
        desc = self.find_text(way_point, "topo:desc")
        src = self.find_text(way_point, "topo:src")
        link = self.parse_link(way_point.find("topo:link", self.name_space))
        sym = self.find_text(way_point, "topo:sym")
        type = self.find_text(way_point, "topo:type")
        fix = self.find_text(way_point, "topo:fix")
        sat = self.find_int(way_point, "topo:sat")
        hdop = self.find_float(way_point, "topo:hdop")
        vdop = self.find_float(way_point, "topo:vdop")
        pdop = self.find_float(way_point, "topo:pdop")
        age_of_gps_data = self.find_float(way_point, "topo:ageofgpsdata")
        dgpsid = self.find_float(way_point, "topo:dgpsid")
        extensions = self.parse_extensions(way_point.find("topo:extensions", self.name_space))

        return WayPoint(tag, lat, lon, ele, time, mag_var, geo_id_height, name, cmt, desc, src, link, sym, type, fix, sat, hdop, vdop, pdop, age_of_gps_data, dgpsid, extensions)

    def find_xmlns_xsi(self) -> Union[str, None]:
        schema_location = None
        for elmt in list(self.xml_root.attrib.keys()):
            if elmt.endswith("schemaLocation"):
                schema_location = elmt[1:-15]
        return schema_location

    def parse_root_properties(self):
        """
        Parse XML properties from GPX file.
        """
        self.gpx.creator = self.xml_root.attrib["creator"]
        self.gpx.version = self.xml_root.attrib["version"]
        self.gpx.xmlns = self.xml_root.tag[1:-4]
        self.gpx.xmlns_xsi = self.find_xmlns_xsi()
        self.name_space["topo"] = self.gpx.xmlns
        name_spaces = self.xml_root.get("{http://www.w3.org/2001/XMLSchema-instance}schemaLocation").split(" ")
        self.gpx.xsi_schema_location = [x for x in name_spaces if x != ""]

    def parse_root_metadata(self):
        """
        Parse metadataType elements from GPX file.
        """
        self.gpx.metadata = self.parse_metadata(self.xml_root.find("topo:metadata", self.name_space))

    def parse_root_way_points(self):
        """
        Parse wptType elements from GPX file.
        """
        way_points = self.xml_root.findall("topo:wpt", self.name_space)
        for way_point in way_points:
            self.gpx.wpt.append(self.parse_way_point(way_point))

    def parse_root_routes(self):
        """
        Parse rteType elements from GPX file
        """
        routes = self.xml_root.findall("topo:rte", self.name_space)
        for route in routes:
            self.gpx.rte.append(self.parse_route(route))  

    def parse_root_tracks(self):
        """
        Parse trkType elements from GPX file.
        """
        tracks = self.xml_root.findall("topo:trk", self.name_space)
        for track in tracks:
            self.gpx.tracks.append(self.parse_track(track))

    def parse_root_extensions(self):
        """
        Parse extensionsType elements from GPX file.
        """
        extensions = self.xml_root.find("topo:extensions", self.name_space)
        self.gpx.extensions = self.parse_extensions(extensions)

    def parse(self) -> Gpx:
        """
        Parse GPX file.

        Returns:
            Gpx: Gpx instance.
        """
        # Parse GPX file
        try:
            self.xml_tree = ET.parse(self.file_path)
            self.xml_root = self.xml_tree.getroot()
        except Exception as err:
            logging.exception(f"Unexpected {err}, {type(err)}.\nUnable to parse GPX file.")
            raise

        # Parse properties
        try:
            self.parse_root_properties()
        except:
            logging.error("Unable to parse properties in GPX file.")
            raise

        # Check XML schemas
        self.check_xml_schemas()

        # Find precisions
        self.find_precisions()

        # Find time format
        self.find_time_format()
        
        # Parse metadata
        try:
            self.parse_root_metadata()
        except:
            logging.error("Unable to parse metadata in GPX file.")
            raise

        # Parse way points
        try:
            self.parse_root_way_points()
        except:
            logging.error("Unable to parse way_points in GPX file.")
            raise

        # Parse routes
        try:
            self.parse_root_routes()
        except:
            logging.error("Unable to parse routes in GPX file.")
            raise

        # Parse tracks
        try:
            self.parse_root_tracks()
        except:
            logging.error("Unable to parse tracks in GPX file.")
            raise

        # Parse extensions
        try:
            self.parse_root_extensions()
        except:
            logging.error("Unable to parse extensions in GPX file.")
            raise

        logging.debug("Parsing complete!!")
        return self.gpx


from typing import Optional, Union
import logging

from ..gpx_elements import Gpx

DEFAULT_PRECISION = 10
DEFAULT_TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

class Parser():
    """
    File parser.
    """

    def __init__(self, file_path: Optional[str] = None) -> None:
        """
        Initialize Parser instance.

        Args:
            file_path (str, optional): Path to the file to parse. Defaults to None.
        """
        self.file_path: str = file_path

        self.precisions: dict = {
            "lat_lon": DEFAULT_PRECISION,
            "elevation": DEFAULT_PRECISION,
            "distance": DEFAULT_PRECISION,
            "duration": DEFAULT_PRECISION,
            "speed": DEFAULT_PRECISION,
            "rate": DEFAULT_PRECISION,
            "default": DEFAULT_PRECISION
            }
        self.time_format = DEFAULT_TIME_FORMAT

        self.gpx: Gpx = Gpx()

    def find_precision(self, number: Union[int, float, str]) -> int:
        """
        Find decimal precision of a given number.

        Parameters
        ----------
        number : Union[int, float, str]
            Number.

        Returns
        -------
        int
            Decimal precision.
        """
        if number is None:
            return DEFAULT_PRECISION
        
        if type(number) == int:
            return 0
        elif type(number) == float:
            number = str(number)
        elif type(number) == str:
            try:
                test = float(number)
            except OSError as err:
                logging.exception(f"OS error: {err}")
            except ValueError:
                logging.exception("Could not convert data ({number}) to a floating point value.")
            except Exception as err:
                logging.exception(f"Unexpected {err}, {type(err)}.\nUnable to find precision of number: {number}")
                raise

        if "." in number:
            _, decimal = number.split(sep=".")
            return len(decimal)
        else:
            return 0


from typing import List

from .extensions import Extensions
from .way_point import WayPoint

class TrackSegment():
    """
    trksegType in GPX file.
    """
    
    def __init__(
            self,
            tag: str = "trkseg",
            trkpt: List[WayPoint] = None,
            extensions: Extensions = None) -> None:
        """
        Initialize TrackSegment instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "trkseg"
        trkpt : List[WayPoint], optional
            List of track points, by default None
        extensions : Extensions, optional
            Extensions, by default None
        """
        self.tag: str = tag
        self.trkpt: List[WayPoint] = [] if trkpt is None else trkpt
        self.extensions: Extensions = extensions

    def project(self, projection: str):
        """
        Project points.

        Parameters
        ----------
        projection : str
            Projection.
        """
        for point in self.trkpt:
            point.project(projection)


import datetime

from ..utils import web_mercator_projection

class Point():
    """
    ptType element in GPX file.
    """

    def __init__(
            self,
            tag: str = "pt",
            lat: float = None,
            lon: float = None,
            ele: float = None,
            time: datetime = None) -> None:
        """
        Initialize Point instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "pt"
        lat : float, optional
            Latitude, by default None
        lon : float, optional
            Longitude, by default None
        ele : float, optional
            Elevation, by default None
        time : datetime, optional
            Time, by default None
        """
        self.tag: str = tag
        self.lat: float = lat
        self.lon: float = lon
        self.ele: float = ele
        self.time: datetime = time

        self._x: int = None
        self._y: int = None

    def project(self):
        self._x, self._y = web_mercator_projection(self.lat, self.lon)

    


from .email import Email
from .link import Link

class Person():
    """
    personType element in GPX file.
    """

    def __init__(
            self,
            tag: str = "person",
            name: str = None,
            email: Email = None,
            link: Link = None) -> None:
        """
        Initialize Person instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "person"
        name : str, optional
            Name, by default None
        email : Email, optional
            Email, by default None
        link : Link, optional
            Link, by default None
        """
        self.tag: str = tag
        self.name: str = name
        self.email: Email = email
        self.link: Link = link